# coding: utf8

# Copyright 2014-2015 Vincent Jacques <vincent@vincent-jacques.net>

import LowVoltage as _lv
import LowVoltage.testing as _tst
from .iterator import Iterator


class BatchGetItemIterator(Iterator):
    """Make as many "BatchGetItem" actions as needed to iterate over all specified items. Including UnprocessedKeys.

    Warning: items are returned by DynamoDB in an unspecified order and LowVoltage does not try to change that.
    """

    def __init__(self, connection, table, *keys):
        self.__table = table
        self.__keys = []
        for key in keys:
            if isinstance(key, dict):
                key = [key]
            self.__keys.extend(key)
        self.__unprocessed_keys = []
        Iterator.__init__(self, connection, self.__next_action())

    def __next_action(self):
        if len(self.__keys) > 0:
            action = _lv.BatchGetItem().table(self.__table).keys(self.__keys[:100])
            self.__keys = self.__keys[100:]
            return action
        elif len(self.__unprocessed_keys) > 0:
            action = _lv.BatchGetItem({self.__table: {"Keys": self.__unprocessed_keys[:100]}})
            self.__unprocessed_keys = self.__unprocessed_keys[100:]
            return action
        else:
            return None

    def process(self, action, r):
        if isinstance(r.unprocessed_keys, dict) and self.__table in r.unprocessed_keys and "Keys" in r.unprocessed_keys[self.__table]:
            self.__unprocessed_keys += r.unprocessed_keys[self.__table]["Keys"]
        return self.__next_action(), r.responses[self.__table]


class BatchGetItemIteratorUnitTests(_tst.UnitTestsWithMocks):
    def setUp(self):
        super(BatchGetItemIteratorUnitTests, self).setUp()
        self.connection = self.mocks.create("connection")

    def test_no_keys(self):
        self.assertEqual(
            list(_lv.BatchGetItemIterator(self.connection.object, "Aaa", [])),
            []
        )

    def test_one_page(self):
        self.connection.expect._call_.withArguments(
            self.ActionChecker("BatchGetItem", {"RequestItems": {"Aaa": {"Keys": [{"h": {"S": "a"}}, {"h": {"S": "b"}}]}}})
        ).andReturn(
            _lv.BatchGetItemResponse(Responses={"Aaa": [{"h": {"S": "c"}}, {"h": {"S": "d"}}]})
        )

        self.assertEqual(
            list(_lv.BatchGetItemIterator(self.connection.object, "Aaa", {"h": u"a"}, {"h": u"b"})),
            [{"h": "c"}, {"h": "d"}]
        )

    def test_one_unprocessed_key(self):
        self.connection.expect._call_.withArguments(
            self.ActionChecker("BatchGetItem", {"RequestItems": {"Aaa": {"Keys": [{"h": {"S": "a"}}, {"h": {"S": "b"}}]}}})
        ).andReturn(
            _lv.BatchGetItemResponse(Responses={"Aaa": [{"h": {"S": "c"}}]}, UnprocessedKeys={"Aaa": {"Keys": [{"h": {"S": "d"}}]}})
        )
        self.connection.expect._call_.withArguments(
            self.ActionChecker("BatchGetItem", {"RequestItems": {"Aaa": {"Keys": [{"h": {"S": "d"}}]}}})
        ).andReturn(
            _lv.BatchGetItemResponse(Responses={"Aaa": [{"h": {"S": "e"}}]})
        )

        self.assertEqual(
            list(_lv.BatchGetItemIterator(self.connection.object, "Aaa", {"h": u"a"}, {"h": u"b"})),
            [{"h": "c"}, {"h": "e"}]
        )

    def test_several_pages(self):
        self.connection.expect._call_.withArguments(
            self.ActionChecker("BatchGetItem", {"RequestItems": {"Aaa": {"Keys": [{"h": {"N": str(i)}} for i in range(0, 100)]}}})
        ).andReturn(
            _lv.BatchGetItemResponse(Responses={"Aaa": [{"h": {"N": str(i)}} for i in range(1000, 1100)]})
        )
        self.connection.expect._call_.withArguments(
            self.ActionChecker("BatchGetItem", {"RequestItems": {"Aaa": {"Keys": [{"h": {"N": str(i)}} for i in range(100, 200)]}}})
        ).andReturn(
            _lv.BatchGetItemResponse(Responses={"Aaa": [{"h": {"N": str(i)}} for i in range(1100, 1200)]})
        )
        self.connection.expect._call_.withArguments(
            self.ActionChecker("BatchGetItem", {"RequestItems": {"Aaa": {"Keys": [{"h": {"N": str(i)}} for i in range(200, 250)]}}})
        ).andReturn(
            _lv.BatchGetItemResponse(Responses={"Aaa": [{"h": {"N": str(i)}} for i in range(1200, 1250)]})
        )

        self.assertEqual(
            list(_lv.BatchGetItemIterator(self.connection.object, "Aaa", ({"h": i} for i in range(0, 250)))),
            [{"h": i} for i in range(1000, 1250)]
        )

    def test_several_pages_of_unprocessed_keys(self):
        self.connection.expect._call_.withArguments(
            self.ActionChecker("BatchGetItem", {"RequestItems": {"Aaa": {"Keys": [{"h": {"N": str(i)}} for i in range(0, 100)]}}})
        ).andReturn(
            _lv.BatchGetItemResponse(
                Responses={"Aaa": [{"h": {"N": str(i)}} for i in range(1000, 1100)]},
                UnprocessedKeys={"Aaa": {"Keys": [{"h": {"N": str(i)}} for i in range(2000, 2075)]}}
            )
        )
        self.connection.expect._call_.withArguments(
            self.ActionChecker("BatchGetItem", {"RequestItems": {"Aaa": {"Keys": [{"h": {"N": str(i)}} for i in range(100, 150)]}}})
        ).andReturn(
            _lv.BatchGetItemResponse(
                Responses={"Aaa": [{"h": {"N": str(i)}} for i in range(1100, 1150)]},
                UnprocessedKeys={"Aaa": {"Keys": [{"h": {"N": str(i)}} for i in range(2075, 2150)]}}
            )
        )
        self.connection.expect._call_.withArguments(
            self.ActionChecker("BatchGetItem", {"RequestItems": {"Aaa": {"Keys": [{"h": {"N": str(i)}} for i in range(2000, 2100)]}}})
        ).andReturn(
            _lv.BatchGetItemResponse(
                Responses={"Aaa": [{"h": {"N": str(i)}} for i in range(1150, 1200)]},
                UnprocessedKeys={"Aaa": {"Keys": [{"h": {"N": str(i)}} for i in range(2150, 2175)]}}
            )
        )
        self.connection.expect._call_.withArguments(
            self.ActionChecker("BatchGetItem", {"RequestItems": {"Aaa": {"Keys": [{"h": {"N": str(i)}} for i in range(2100, 2175)]}}})
        ).andReturn(
            _lv.BatchGetItemResponse(
                Responses={"Aaa": [{"h": {"N": str(i)}} for i in range(1200, 1250)]},
            )
        )

        self.assertEqual(
            list(_lv.BatchGetItemIterator(self.connection.object, "Aaa", [{"h": i} for i in range(0, 150)])),
            [{"h": i} for i in range(1000, 1250)]
        )


class BatchGetItemIteratorLocalIntegTests(_tst.LocalIntegTestsWithTableH):
    def key(self, i):
        return u"{:03}".format(i)

    def setUp(self):
        super(BatchGetItemIteratorLocalIntegTests, self).setUp()
        _lv.BatchPutItem(self.connection, "Aaa", [{"h": self.key(i), "xs": "x" * 300000} for i in range(250)])  # 300kB items ensure a single BatchGetItem will return at most 55 items

    def test(self):
        keys = [item["h"] for item in _lv.BatchGetItemIterator(self.connection, "Aaa", ({"h": self.key(i)} for i in range(250)))]
        self.assertEqual(sorted(keys), [self.key(i) for i in range(250)])



# coding: utf8

# Copyright 2014-2015 Vincent Jacques <vincent@vincent-jacques.net>

import LowVoltage as _lv
import LowVoltage.testing as _tst


def BatchDeleteItem(connection, table, *keys):
    """Make as many "BatchWriteItem" actions as needed to delete all specified keys. Including UnprocessedItems."""

    delete = []
    for key in keys:
        if isinstance(key, dict):
            key = [key]
        delete.extend(key)

    unprocessed_items = []

    while len(delete) != 0:
        r = connection(_lv.BatchWriteItem().table(table).delete(delete[:25]))
        delete = delete[25:]
        if isinstance(r.unprocessed_items, dict) and table in r.unprocessed_items:
            unprocessed_items += r.unprocessed_items[table]

    # @todo Maybe wait a bit before retrying unprocessed items? Same in BatchPutItem and BatchGetItemIterator.
    # @todo In the first loop, maybe wait a bit before next request if we get unprocessed items? Might not be a good idea.

    while len(unprocessed_items) != 0:
        r = connection(_lv.BatchWriteItem({table: unprocessed_items[:25]}))
        unprocessed_items = unprocessed_items[25:]
        if isinstance(r.unprocessed_items, dict) and table in r.unprocessed_items:
            unprocessed_items += r.unprocessed_items[table]


class BatchDeleteItemUnitTests(_tst.UnitTestsWithMocks):
    def setUp(self):
        super(BatchDeleteItemUnitTests, self).setUp()
        self.connection = self.mocks.create("connection")

    def test_no_keys(self):
        BatchDeleteItem(self.connection.object, "Aaa", [])

    def test_one_page(self):
        self.connection.expect._call_.withArguments(
            self.ActionChecker("BatchWriteItem", {"RequestItems": {"Aaa": [{"DeleteRequest": {"Key": {"h": {"S": "a"}}}}, {"DeleteRequest": {"Key": {"h": {"S": "b"}}}}]}})
        ).andReturn(
            _lv.BatchWriteItemResponse()
        )

        BatchDeleteItem(self.connection.object, "Aaa", {"h": u"a"}, {"h": u"b"})

    def test_several_pages(self):
        self.connection.expect._call_.withArguments(
            self.ActionChecker("BatchWriteItem", {"RequestItems": {"Aaa": [{"DeleteRequest": {"Key": {"h": {"N": str(i)}}}} for i in range(0, 25)]}})
        ).andReturn(
            _lv.BatchWriteItemResponse()
        )
        self.connection.expect._call_.withArguments(
            self.ActionChecker("BatchWriteItem", {"RequestItems": {"Aaa": [{"DeleteRequest": {"Key": {"h": {"N": str(i)}}}} for i in range(25, 50)]}})
        ).andReturn(
            _lv.BatchWriteItemResponse()
        )
        self.connection.expect._call_.withArguments(
            self.ActionChecker("BatchWriteItem", {"RequestItems": {"Aaa": [{"DeleteRequest": {"Key": {"h": {"N": str(i)}}}} for i in range(50, 60)]}})
        ).andReturn(
            _lv.BatchWriteItemResponse()
        )

        BatchDeleteItem(self.connection.object, "Aaa", ({"h": i} for i in range(60)))

    def test_one_unprocessed_item(self):
        self.connection.expect._call_.withArguments(
            self.ActionChecker("BatchWriteItem", {"RequestItems": {"Aaa": [{"DeleteRequest": {"Key": {"h": {"S": "a"}}}}, {"DeleteRequest": {"Key": {"h": {"S": "b"}}}}]}})
        ).andReturn(
            _lv.BatchWriteItemResponse(UnprocessedItems={"Aaa": [{"DeleteRequest": {"Key": {"h": {"S": "c"}}}}]})
        )
        self.connection.expect._call_.withArguments(
            self.ActionChecker("BatchWriteItem", {"RequestItems": {"Aaa": [{"DeleteRequest": {"Key": {"h": {"S": "c"}}}}]}})
        ).andReturn(
            _lv.BatchWriteItemResponse()
        )

        BatchDeleteItem(self.connection.object, "Aaa", {"h": u"a"}, {"h": u"b"})

    def test_several_pages_of_unprocessed_item(self):
        self.connection.expect._call_.withArguments(
            self.ActionChecker("BatchWriteItem", {"RequestItems": {"Aaa": [{"DeleteRequest": {"Key": {"h": {"N": str(i)}}}} for i in range(0, 25)]}})
        ).andReturn(
            _lv.BatchWriteItemResponse(UnprocessedItems={"Aaa": [{"DeleteRequest": {"Key": {"h": {"N": str(i)}}}} for i in range(100, 110)]})
        )
        self.connection.expect._call_.withArguments(
            self.ActionChecker("BatchWriteItem", {"RequestItems": {"Aaa": [{"DeleteRequest": {"Key": {"h": {"N": str(i)}}}} for i in range(25, 35)]}})
        ).andReturn(
            _lv.BatchWriteItemResponse(UnprocessedItems={"Aaa": [{"DeleteRequest": {"Key": {"h": {"N": str(i)}}}} for i in range(110, 120)]})
        )
        self.connection.expect._call_.withArguments(
            self.ActionChecker("BatchWriteItem", {"RequestItems": {"Aaa": [{"DeleteRequest": {"Key": {"h": {"N": str(i)}}}} for i in range(100, 120)]}})
        ).andReturn(
            _lv.BatchWriteItemResponse(UnprocessedItems={"Aaa": [{"DeleteRequest": {"Key": {"h": {"N": str(i)}}}} for i in range(120, 130)]})
        )
        self.connection.expect._call_.withArguments(
            self.ActionChecker("BatchWriteItem", {"RequestItems": {"Aaa": [{"DeleteRequest": {"Key": {"h": {"N": str(i)}}}} for i in range(120, 130)]}})
        ).andReturn(
            _lv.BatchWriteItemResponse()
        )

        BatchDeleteItem(self.connection.object, "Aaa", [{"h": i} for i in range(35)])


class BatchDeleteItemLocalIntegTests(_tst.LocalIntegTestsWithTableH):
    def key(self, i):
        return u"{:03}".format(i)

    def setUp(self):
        super(BatchDeleteItemLocalIntegTests, self).setUp()
        _lv.BatchPutItem(self.connection, "Aaa", [{"h": self.key(i)} for i in range(100)])

    def test(self):
        _lv.BatchDeleteItem(self.connection, "Aaa", [{"h": self.key(i)} for i in range(100)])
        self.assertEqual([], list(_lv.ScanIterator(self.connection, _lv.Scan("Aaa"))))



# coding: utf8

# Copyright 2014-2015 Vincent Jacques <vincent@vincent-jacques.net>

"""
When given a :class:`CreateTable`, the connection will return a :class:`CreateTableResponse`:

.. testsetup::

    table = "LowVoltage.Tests.Doc.CreateTable.1"
    table2 = "LowVoltage.Tests.Doc.CreateTable.2"
    table3 = "LowVoltage.Tests.Doc.CreateTable.3"
    table4 = "LowVoltage.Tests.Doc.CreateTable.4"

>>> r = connection(
...   CreateTable(table)
...     .hash_key("h", STRING)
...     .provisioned_throughput(1, 1)
... )
>>> r
<LowVoltage.actions.create_table.CreateTableResponse object at ...>
>>> r.table_description.table_status
u'CREATING'

Note that you can use :func:`.WaitForTableActivation` to poll the table status until it's usable.

.. testcleanup::

    WaitForTableActivation(connection, table)
    WaitForTableActivation(connection, table2)
    WaitForTableActivation(connection, table3)
    WaitForTableActivation(connection, table4)
    connection(DeleteTable(table))
    connection(DeleteTable(table2))
    connection(DeleteTable(table3))
    connection(DeleteTable(table4))
    WaitForTableDeletion(connection, table)
    WaitForTableDeletion(connection, table2)
    WaitForTableDeletion(connection, table3)
    WaitForTableDeletion(connection, table4)
"""

import datetime

import LowVoltage as _lv
import LowVoltage.testing as _tst
from .action import Action
from .return_types import TableDescription, _is_dict


class CreateTableResponse(object):
    """
    The `CreateTable response <http://docs.aws.amazon.com/amazondynamodb/latest/APIReference/API_CreateTable.html#API_CreateTable_ResponseElements>`__.
    """

    def __init__(
        self,
        TableDescription=None,
        **dummy
    ):
        self.__table_description = TableDescription

    @property
    def table_description(self):
        """
        The description of the table you just created.

        :type: ``None`` or :class:`.TableDescription`
        """
        if _is_dict(self.__table_description):  # pragma no branch (Defensive code)
            return TableDescription(**self.__table_description)


class CreateTable(Action):
    """
    The `CreateTable request <http://docs.aws.amazon.com/amazondynamodb/latest/APIReference/API_CreateTable.html#API_CreateTable_RequestParameters>`__.
    """

    # @todo Should we add ctor parameters and allow use to choose between ctor and builder syntaxes? Same for .global_secondary_index. Same everywhere.

    def __init__(self, table_name):
        super(CreateTable, self).__init__("CreateTable", CreateTableResponse)
        self.__table_name = table_name
        self._hash_key = None
        self._range_key = None
        self.__attribute_definitions = {}
        self._read_capacity_units = None
        self._write_capacity_units = None
        self.__active_index = self
        self.__gsis = {}
        self.__lsis = {}

    @property
    def payload(self):
        data = {"TableName": self.__table_name}
        schema = []
        if self._hash_key:
            schema.append({"AttributeName": self._hash_key, "KeyType": "HASH"})
        if self._range_key:
            schema.append({"AttributeName": self._range_key, "KeyType": "RANGE"})
        if schema:
            data["KeySchema"] = schema
        if self.__attribute_definitions:
            data["AttributeDefinitions"] = [
                {"AttributeName": name, "AttributeType": typ}
                for name, typ in self.__attribute_definitions.iteritems()
            ]
        throughput = {}
        if self._read_capacity_units:
            throughput["ReadCapacityUnits"] = self._read_capacity_units
        if self._write_capacity_units:
            throughput["WriteCapacityUnits"] = self._write_capacity_units
        if throughput:
            data["ProvisionedThroughput"] = throughput
        if self.__gsis:
            data["GlobalSecondaryIndexes"] = [i.payload for i in self.__gsis.itervalues()]
        if self.__lsis:
            data["LocalSecondaryIndexes"] = [i.payload for i in self.__lsis.itervalues()]
        return data

    class _Index(object):
        def __init__(self, name):
            self.__name = name
            self._hash_key = None
            self._range_key = None
            self._projection = None

        @property
        def payload(self):
            data = {"IndexName": self.__name}
            schema = []
            if self._hash_key:
                schema.append({"AttributeName": self._hash_key, "KeyType": "HASH"})
            if self._range_key:
                schema.append({"AttributeName": self._range_key, "KeyType": "RANGE"})
            if schema:
                data["KeySchema"] = schema
            if isinstance(self._projection, basestring):
                data["Projection"] = {"ProjectionType": self._projection}
            elif self._projection:
                data["Projection"] = {"ProjectionType": "INCLUDE", "NonKeyAttributes": self._projection}
            return data

    class _IndexWithThroughput(_Index):
        def __init__(self, name):
            super(CreateTable._IndexWithThroughput, self).__init__(name)
            self._read_capacity_units = None
            self._write_capacity_units = None

        @property
        def payload(self):
            data = super(CreateTable._IndexWithThroughput, self).payload
            throughput = {}
            if self._read_capacity_units:
                throughput["ReadCapacityUnits"] = self._read_capacity_units
            if self._write_capacity_units:
                throughput["WriteCapacityUnits"] = self._write_capacity_units
            if throughput:
                data["ProvisionedThroughput"] = throughput
            return data

    def hash_key(self, name, typ=None):
        """
        Set the hash key in KeySchema for the table or the active index.
        If you provide a second argument, :meth:`attribute_definition` will be called as well.

        See :meth:`range_key` for examples.
        """
        self.__active_index._hash_key = name
        if typ is not None:
            self.attribute_definition(name, typ)
        return self

    def range_key(self, name, typ=None):
        """
        Set the range key in KeySchema for the table or the active index.
        If you provide a second argument, :meth:`attribute_definition` will be called as well.

        >>> connection(
        ...   CreateTable(table2)
        ...     .hash_key("h", STRING)
        ...     .range_key("r")
        ...     .provisioned_throughput(1, 1)
        ...     .attribute_definition("r", NUMBER)
        ... )
        <LowVoltage.actions.create_table.CreateTableResponse ...>
        """
        self.__active_index._range_key = name
        if typ is not None:
            self.attribute_definition(name, typ)
        return self

    def attribute_definition(self, name, typ):
        """
        Set the type of an attribute in AttributeDefinitions.
        Key attribute must be typed. See :mod:`.attribute_types` for constants to be passed to this method.

        See :meth:`range_key` for an example.
        """
        self.__attribute_definitions[name] = typ
        return self

    def provisioned_throughput(self, read_capacity_units, write_capacity_units):
        """
        Set the read and write provisioned throughput for the table or the active index.

        See :meth:`range_key`, :meth:`global_secondary_index` or :meth:`local_secondary_index` for examples.
        """
        self.__active_index._read_capacity_units = read_capacity_units
        self.__active_index._write_capacity_units = write_capacity_units
        return self

    def global_secondary_index(self, name):
        """
        Add a GSI. This method sets the active index: methods like :meth:`hash_key` will apply to the index.

        >>> connection(
        ...   CreateTable(table3)
        ...     .hash_key("h", STRING)
        ...     .provisioned_throughput(1, 1)
        ...     .global_secondary_index("gsi")
        ...       .hash_key("a", BINARY)
        ...       .range_key("b", NUMBER)
        ...       .provisioned_throughput(1, 1)
        ...       .project_all()
        ... )
        <LowVoltage.actions.create_table.CreateTableResponse ...>
        """
        if name not in self.__gsis:
            self.__gsis[name] = self._IndexWithThroughput(name)
        self.__active_index = self.__gsis[name]
        return self

    def local_secondary_index(self, name):
        """
        Add a LSI. This method sets the active index: methods like :meth:`hash_key` will apply to the index.

        >>> connection(
        ...   CreateTable(table4)
        ...     .hash_key("h", STRING)
        ...     .range_key("r", NUMBER)
        ...     .provisioned_throughput(1, 1)
        ...     .local_secondary_index("lsi")
        ...       .hash_key("h")
        ...       .range_key("a", NUMBER)
        ...       .provisioned_throughput(1, 1)
        ...       .project("x", "y")
        ... )
        <LowVoltage.actions.create_table.CreateTableResponse ...>
        """
        if name not in self.__lsis:
            self.__lsis[name] = self._Index(name)
        # @todo Should we copy the hash key from the table? (It must be the same)
        self.__active_index = self.__lsis[name]
        return self

    def table(self):
        """
        Reset the active index: methods like :meth:`hash_key` will apply to the table.
        """
        self.__active_index = self
        return self

    def project_all(self):
        """
        Set ProjectionType to ALL for the active index.

        :raise: :exc:`.BuilderError` if called when no index is active.

        See :meth:`global_secondary_index` for an example.
        """
        self.__check_active_index()
        self.__active_index._projection = "ALL"
        return self

    def project_keys_only(self):
        """
        Set ProjectionType to KEYS_ONLY for the active index.

        :raise: :exc:`.BuilderError` if called when no index is active.
        """
        self.__check_active_index()
        self.__active_index._projection = "KEYS_ONLY"
        return self

    def project(self, *attrs):
        """
        Set ProjectionType to INCLUDE for the active index and add names to NonKeyAttributes.

        :raise: :exc:`.BuilderError` if called when no index is active.

        See :meth:`local_secondary_index` for an example.
        """
        self.__check_active_index()
        if not isinstance(self.__active_index._projection, list):
            self.__active_index._projection = []
        for attr in attrs:
            if isinstance(attr, basestring):
                attr = [attr]
            self.__active_index._projection.extend(attr)
        return self

    def __check_active_index(self):
        if self.__active_index is self:
            raise _lv.BuilderError("No active index.")


class CreateTableUnitTests(_tst.UnitTests):
    def test_name(self):
        self.assertEqual(CreateTable("Foo").name, "CreateTable")

    def test_no_arguments(self):
        self.assertEqual(CreateTable("Foo").payload, {"TableName": "Foo"})

    def test_hash_key(self):
        self.assertEqual(
            CreateTable("Foo").hash_key("h").payload,
            {
                "TableName": "Foo",
                "KeySchema": [{"AttributeName": "h", "KeyType": "HASH"}],
            }
        )

    def test_hash_key_with_type(self):
        self.assertEqual(
            CreateTable("Foo").hash_key("h", _lv.STRING).payload,
            {
                "TableName": "Foo",
                "AttributeDefinitions": [{"AttributeName": "h", "AttributeType": "S"}],
                "KeySchema": [{"AttributeName": "h", "KeyType": "HASH"}],
            }
        )

    def test_attribute_definition(self):
        self.assertEqual(
            CreateTable("Foo").attribute_definition("h", _lv.STRING).payload,
            {
                "TableName": "Foo",
                "AttributeDefinitions": [{"AttributeName": "h", "AttributeType": "S"}],
            }
        )

    def test_range_key(self):
        self.assertEqual(
            CreateTable("Foo").range_key("r").payload,
            {
                "TableName": "Foo",
                "KeySchema": [{"AttributeName": "r", "KeyType": "RANGE"}],
            }
        )

    def test_range_key_with_type(self):
        self.assertEqual(
            CreateTable("Foo").range_key("r", _lv.STRING).payload,
            {
                "TableName": "Foo",
                "AttributeDefinitions": [{"AttributeName": "r", "AttributeType": "S"}],
                "KeySchema": [{"AttributeName": "r", "KeyType": "RANGE"}],
            }
        )

    def test_throughput(self):
        self.assertEqual(
            CreateTable("Foo").provisioned_throughput(42, 43).payload,
            {
                "TableName": "Foo",
                "ProvisionedThroughput": {"ReadCapacityUnits": 42, "WriteCapacityUnits": 43},
            }
        )

    def test_global_secondary_index(self):
        self.assertEqual(
            CreateTable("Foo").global_secondary_index("foo").payload,
            {
                "TableName": "Foo",
                "GlobalSecondaryIndexes": [{"IndexName": "foo"}],
            }
        )

    def test_local_secondary_index(self):
        self.assertEqual(
            CreateTable("Foo").local_secondary_index("foo").payload,
            {
                "TableName": "Foo",
                "LocalSecondaryIndexes": [{"IndexName": "foo"}],
            }
        )

    def test_global_secondary_index_hash_key(self):
        self.assertEqual(
            CreateTable("Foo").global_secondary_index("foo").hash_key("hh").payload,
            {
                "TableName": "Foo",
                "GlobalSecondaryIndexes": [
                    {"IndexName": "foo", "KeySchema": [{"AttributeName": "hh", "KeyType": "HASH"}]},
                ],
            }
        )

    def test_global_secondary_index_range_key(self):
        self.assertEqual(
            CreateTable("Foo").global_secondary_index("foo").range_key("rr").payload,
            {
                "TableName": "Foo",
                "GlobalSecondaryIndexes": [
                    {"IndexName": "foo", "KeySchema": [{"AttributeName": "rr", "KeyType": "RANGE"}]},
                ],
            }
        )

    def test_global_secondary_index_hash_key_with_type(self):
        self.assertEqual(
            CreateTable("Foo").global_secondary_index("foo").hash_key("hh", _lv.STRING).payload,
            {
                "TableName": "Foo",
                "GlobalSecondaryIndexes": [
                    {"IndexName": "foo", "KeySchema": [{"AttributeName": "hh", "KeyType": "HASH"}]},
                ],
                "AttributeDefinitions": [{"AttributeName": "hh", "AttributeType": "S"}],
            }
        )

    def test_global_secondary_index_range_key_with_type(self):
        self.assertEqual(
            CreateTable("Foo").global_secondary_index("foo").range_key("rr", _lv.STRING).payload,
            {
                "TableName": "Foo",
                "GlobalSecondaryIndexes": [
                    {"IndexName": "foo", "KeySchema": [{"AttributeName": "rr", "KeyType": "RANGE"}]},
                ],
                "AttributeDefinitions": [{"AttributeName": "rr", "AttributeType": "S"}],
            }
        )

    def test_global_secondary_index_throughput(self):
        self.assertEqual(
            CreateTable("Foo").global_secondary_index("foo").provisioned_throughput(42, 43).payload,
            {
                "TableName": "Foo",
                "GlobalSecondaryIndexes": [
                    {"IndexName": "foo", "ProvisionedThroughput": {"ReadCapacityUnits": 42, "WriteCapacityUnits": 43}},
                ],
            }
        )

    def test_global_secondary_index_project_all(self):
        self.assertEqual(
            CreateTable("Foo").global_secondary_index("foo").project_all().payload,
            {
                "TableName": "Foo",
                "GlobalSecondaryIndexes": [
                    {"IndexName": "foo", "Projection": {"ProjectionType": "ALL"}},
                ],
            }
        )

    def test_global_secondary_index_project_keys_only(self):
        self.assertEqual(
            CreateTable("Foo").global_secondary_index("foo").project_keys_only().payload,
            {
                "TableName": "Foo",
                "GlobalSecondaryIndexes": [
                    {"IndexName": "foo", "Projection": {"ProjectionType": "KEYS_ONLY"}},
                ],
            }
        )

    def test_global_secondary_index_project_include(self):
        self.assertEqual(
            CreateTable("Foo").global_secondary_index("foo").project("toto", "titi").project(["tutu"]).payload,
            {
                "TableName": "Foo",
                "GlobalSecondaryIndexes": [
                    {"IndexName": "foo", "Projection": {"ProjectionType": "INCLUDE", "NonKeyAttributes": ["toto", "titi", "tutu"]}},
                ],
            }
        )

    def test_back_to_table_after_gsi(self):
        self.assertEqual(
            CreateTable("Foo").global_secondary_index("foo").table().provisioned_throughput(42, 43).payload,
            {
                "TableName": "Foo",
                "GlobalSecondaryIndexes": [{"IndexName": "foo"}],
                "ProvisionedThroughput": {"ReadCapacityUnits": 42, "WriteCapacityUnits": 43},
            }
        )

    def test_implicit_back_to_table_after_gsi(self):
        self.assertEqual(
            CreateTable("Foo").global_secondary_index("foo").attribute_definition("bar", _lv.NUMBER).payload,
            {
                "TableName": "Foo",
                "GlobalSecondaryIndexes": [{"IndexName": "foo"}],
                "AttributeDefinitions": [{"AttributeName": "bar", "AttributeType": "N"}],
            }
        )

    def test_back_to_gsi_after_back_to_table(self):
        self.assertEqual(
            CreateTable("Foo")
                .global_secondary_index("foo")
                .table().provisioned_throughput(42, 43)
                .global_secondary_index("foo").provisioned_throughput(12, 13)
                .payload,
            {
                "TableName": "Foo",
                "GlobalSecondaryIndexes": [
                    {"IndexName": "foo", "ProvisionedThroughput": {"ReadCapacityUnits": 12, "WriteCapacityUnits": 13}}
                ],
                "ProvisionedThroughput": {"ReadCapacityUnits": 42, "WriteCapacityUnits": 43},
            }
        )

    def test_back_to_lsi_after_back_to_table(self):
        self.assertEqual(
            CreateTable("Foo")
                .local_secondary_index("foo")
                .table().provisioned_throughput(42, 43)
                .local_secondary_index("foo").project_all()
                .payload,
            {
                "TableName": "Foo",
                "LocalSecondaryIndexes": [
                    {"IndexName": "foo", "Projection": {"ProjectionType": "ALL"}}
                ],
                "ProvisionedThroughput": {"ReadCapacityUnits": 42, "WriteCapacityUnits": 43},
            }
        )

    def test_project_without_active_index(self):
        with self.assertRaises(_lv.BuilderError) as catcher:
            CreateTable("Foo").project("a")
        self.assertEqual(catcher.exception.args, ("No active index.",))

    def test_project_all_without_active_index(self):
        with self.assertRaises(_lv.BuilderError) as catcher:
            CreateTable("Foo").project_all()
        self.assertEqual(catcher.exception.args, ("No active index.",))

    def test_project_keys_only_without_active_index(self):
        with self.assertRaises(_lv.BuilderError) as catcher:
            CreateTable("Foo").project_keys_only()
        self.assertEqual(catcher.exception.args, ("No active index.",))


class CreateTableLocalIntegTests(_tst.LocalIntegTests):
    def tearDown(self):
        self.connection(_lv.DeleteTable("Aaa"))
        super(CreateTableLocalIntegTests, self).tearDown()

    def test_simplest_table(self):
        r = self.connection(
            _lv.CreateTable("Aaa").hash_key("h", _lv.STRING).provisioned_throughput(1, 2)
        )

        self.assertDateTimeIsReasonable(r.table_description.creation_date_time)
        self.assertEqual(r.table_description.attribute_definitions[0].attribute_name, "h")
        self.assertEqual(r.table_description.attribute_definitions[0].attribute_type, "S")
        self.assertEqual(r.table_description.global_secondary_indexes, None)
        self.assertEqual(r.table_description.item_count, 0)
        self.assertEqual(r.table_description.key_schema[0].attribute_name, "h")
        self.assertEqual(r.table_description.key_schema[0].key_type, "HASH")
        self.assertEqual(r.table_description.local_secondary_indexes, None)
        self.assertEqual(r.table_description.provisioned_throughput.last_decrease_date_time, datetime.datetime(1970, 1, 1))
        self.assertEqual(r.table_description.provisioned_throughput.last_increase_date_time, datetime.datetime(1970, 1, 1))
        self.assertEqual(r.table_description.provisioned_throughput.number_of_decreases_today, 0)
        self.assertEqual(r.table_description.provisioned_throughput.read_capacity_units, 1)
        self.assertEqual(r.table_description.provisioned_throughput.write_capacity_units, 2)
        self.assertEqual(r.table_description.table_name, "Aaa")
        self.assertEqual(r.table_description.table_size_bytes, 0)
        self.assertEqual(r.table_description.table_status, "ACTIVE")

    def test_simple_global_secondary_index(self):
        r = self.connection(
            _lv.CreateTable("Aaa").hash_key("h", _lv.STRING).provisioned_throughput(1, 2)
                .global_secondary_index("the_gsi")
                .hash_key("hh", _lv.STRING)
                .project_all()
                .provisioned_throughput(3, 4)
        )

        self.assertEqual(r.table_description.global_secondary_indexes[0].index_name, "the_gsi")
        self.assertEqual(r.table_description.global_secondary_indexes[0].index_size_bytes, 0)
        self.assertEqual(r.table_description.global_secondary_indexes[0].index_status, "ACTIVE")
        self.assertEqual(r.table_description.global_secondary_indexes[0].item_count, 0)
        self.assertEqual(r.table_description.global_secondary_indexes[0].key_schema[0].attribute_name, "hh")
        self.assertEqual(r.table_description.global_secondary_indexes[0].key_schema[0].key_type, "HASH")
        self.assertEqual(r.table_description.global_secondary_indexes[0].projection.non_key_attributes, None)
        self.assertEqual(r.table_description.global_secondary_indexes[0].projection.projection_type, "ALL")
        self.assertEqual(r.table_description.global_secondary_indexes[0].provisioned_throughput.last_decrease_date_time, None)
        self.assertEqual(r.table_description.global_secondary_indexes[0].provisioned_throughput.last_increase_date_time, None)
        self.assertEqual(r.table_description.global_secondary_indexes[0].provisioned_throughput.number_of_decreases_today, None)
        self.assertEqual(r.table_description.global_secondary_indexes[0].provisioned_throughput.read_capacity_units, 3)
        self.assertEqual(r.table_description.global_secondary_indexes[0].provisioned_throughput.write_capacity_units, 4)

    def test_simple_local_secondary_index(self):
        r = self.connection(
            _lv.CreateTable("Aaa").hash_key("h", _lv.STRING).range_key("r", _lv.STRING).provisioned_throughput(1, 2)
                .local_secondary_index("the_lsi").hash_key("h").range_key("rr", _lv.STRING).project_all()
        )

        self.assertEqual(r.table_description.local_secondary_indexes[0].index_name, "the_lsi")
        self.assertEqual(r.table_description.local_secondary_indexes[0].index_size_bytes, 0)
        self.assertEqual(r.table_description.local_secondary_indexes[0].item_count, 0)
        self.assertEqual(r.table_description.local_secondary_indexes[0].key_schema[0].attribute_name, "h")
        self.assertEqual(r.table_description.local_secondary_indexes[0].key_schema[0].key_type, "HASH")
        self.assertEqual(r.table_description.local_secondary_indexes[0].key_schema[1].attribute_name, "rr")
        self.assertEqual(r.table_description.local_secondary_indexes[0].key_schema[1].key_type, "RANGE")
        self.assertEqual(r.table_description.local_secondary_indexes[0].projection.non_key_attributes, None)
        self.assertEqual(r.table_description.local_secondary_indexes[0].projection.projection_type, "ALL")

    def test_global_secondary_index_with_projection(self):
        r = self.connection(
            _lv.CreateTable("Aaa").hash_key("h", _lv.STRING).provisioned_throughput(1, 2)
                .global_secondary_index("the_gsi")
                .hash_key("hh", _lv.STRING)
                .project("toto", "titi")
                .provisioned_throughput(3, 4)
        )

        self.assertEqual(r.table_description.global_secondary_indexes[0].projection.non_key_attributes[0], "toto")
        self.assertEqual(r.table_description.global_secondary_indexes[0].projection.non_key_attributes[1], "titi")
        self.assertEqual(r.table_description.global_secondary_indexes[0].projection.projection_type, "INCLUDE")


class CreateTableErrorLocalIntegTests(_tst.LocalIntegTests):
    def test_define_unused_attribute(self):
        with self.assertRaises(_lv.ValidationException) as catcher:
            self.connection(
                _lv.CreateTable("Aaa").hash_key("h", _lv.STRING).provisioned_throughput(1, 2)
                    .attribute_definition("x", _lv.STRING)
            )
        self.assertEqual(
            catcher.exception.args,
            ({
                "__type": "com.amazon.coral.validate#ValidationException",
                "Message": "The number of attributes in key schema must match the number of attributesdefined in attribute definitions.",
            },)
        )

    def test_dont_define_key_attribute(self):
        with self.assertRaises(_lv.ValidationException) as catcher:
            self.connection(
                _lv.CreateTable("Aaa").hash_key("h").provisioned_throughput(1, 2)
                    .attribute_definition("x", _lv.STRING)
            )
        self.assertEqual(
            catcher.exception.args,
            ({
                "__type": "com.amazon.coral.validate#ValidationException",
                "Message": "Hash Key not specified in Attribute Definitions.  Type unknown.",
            },)
        )

    def test_dont_define_any_attribute(self):
        with self.assertRaises(_lv.ValidationException) as catcher:
            self.connection(
                _lv.CreateTable("Aaa").hash_key("h").provisioned_throughput(1, 2)
            )
        self.assertEqual(
            catcher.exception.args,
            ({
                "__type": "com.amazon.coral.validate#ValidationException",
                "Message": "No Attribute Schema Defined",
            },)
        )



# coding: utf8

# Copyright 2014-2015 Vincent Jacques <vincent@vincent-jacques.net>

"""
@todo Link to user guide (error management).

Some errors are retryable: :exc:`.NetworkError` or :exc:`.ProvisionedThroughputExceededException` for example.
When they happen, it makes sense to wait for a short while and then retry the request.
Retry policies decide if it's worth retrying and how long to wait.

.. py:class:: RetryPolicy

    The interface to be implemented by all retry policies. Note that you must not inherit from this class, just implement the same interface.

    .. py:method:: retry(action, exceptions)

        Return the delay before retrying the action. ``None`` if the action shouldn't be retried.

        :param action: the action that failed.
        :param exceptions: the (retryable) exceptions that occured so far. The most recent exception is ``exceptions[-1]``.

        :type: ``None`` or number (in seconds)
"""

import LowVoltage as _lv
import LowVoltage.testing as _tst


class ExponentialBackoffRetryPolicy(object):
    """
    Retry failed requests with a waiting time growing exponentialy.

    :param first_wait: the duration to wait before the first retry.
    :param multiplier: the factor by which to augment the waiting duration between successive retries. Greater than 1.
    :param max_retries: the maximum number of times to retry a failed action (0 meaning "don't retry", 1 meaning "retry once", etc.).
    """
    def __init__(self, first_wait, multiplier, max_retries):
        self.__first_wait = first_wait
        self.__multiplier = multiplier
        self.__max_retries = max_retries

    def retry(self, action, exceptions):
        # @todo Should we wait different times for different errors?
        if len(exceptions) > self.__max_retries:
            return None
        else:
            return self.__first_wait * (self.__multiplier ** (len(exceptions) - 1))

    # This is needed for the documentation of DEFAULT
    def __repr__(self):
        return "ExponentialBackoffRetryPolicy({}, {}, {})".format(self.__first_wait, self.__multiplier, self.__max_retries)


DEFAULT = ExponentialBackoffRetryPolicy(1, 1.5, 4)
"""
The default retry policy: a reasonable exponential backoff.
"""


FAIL_FAST = ExponentialBackoffRetryPolicy(1, 1, 0)
"""
A retry policy that never retries anything.
"""


class ExponentialBackoffRetryPolicyUnitTests(_tst.UnitTests):
    def setUp(self):
        self.policy = ExponentialBackoffRetryPolicy(1, 3, 3)
        super(ExponentialBackoffRetryPolicyUnitTests, self).setUp()

    def test_wait_after_first_failure(self):
        self.assertEqual(
            self.policy.retry(object(), [_lv.ServerError()]),
            1
        )

    def test_wait_after_second_failure(self):
        self.assertEqual(
            self.policy.retry(object(), [_lv.ServerError(), _lv.ServerError()]),
            3
        )

    def test_wait_after_third_failure(self):
        self.assertEqual(
            self.policy.retry(object(), [_lv.ServerError(), _lv.ServerError(), _lv.ServerError()]),
            9
        )

    def test_give_up_after_fourth_failure(self):
        self.assertIsNone(self.policy.retry(object(), [_lv.ServerError(), _lv.ServerError(), _lv.ServerError(), _lv.ServerError()]))

    def test_repr(self):
        self.assertEqual(repr(self.policy), "ExponentialBackoffRetryPolicy(1, 3, 3)")



# coding: utf8

# Copyright 2014-2015 Vincent Jacques <vincent@vincent-jacques.net>

from .batch_get_item import BatchGetItem, BatchGetItemResponse
from .batch_write_item import BatchWriteItem, BatchWriteItemResponse
from .create_table import CreateTable, CreateTableResponse
from .delete_item import DeleteItem, DeleteItemResponse
from .delete_table import DeleteTable, DeleteTableResponse
from .describe_table import DescribeTable, DescribeTableResponse
from .get_item import GetItem, GetItemResponse
from .list_tables import ListTables, ListTablesResponse
from .put_item import PutItem, PutItemResponse
from .query import Query, QueryResponse
from .scan import Scan, ScanResponse
from .update_item import UpdateItem, UpdateItemResponse
from .update_table import UpdateTable, UpdateTableResponse



# coding: utf8

# Copyright 2014-2015 Vincent Jacques <vincent@vincent-jacques.net>

import LowVoltage as _lv
import LowVoltage.testing as _tst
from .iterator import Iterator


class ListTablesIterator(Iterator):
    """Make as many "ListTables" actions as needed to iterate over all tables."""

    def __init__(self, connection):
        Iterator.__init__(self, connection, _lv.ListTables())

    def process(self, action, r):
        if r.last_evaluated_table_name is None:
            action = None
        else:
            action.exclusive_start_table_name(r.last_evaluated_table_name)
        return action, r.table_names


class ListTablesIteratorUnitTests(_tst.UnitTestsWithMocks):
    def setUp(self):
        super(ListTablesIteratorUnitTests, self).setUp()
        self.connection = self.mocks.create("connection")

    def test_no_tables(self):
        self.connection.expect._call_.withArguments(self.ActionChecker("ListTables", {})).andReturn(_lv.ListTablesResponse(TableNames=[]))

        self.assertEqual(
            list(ListTablesIterator(self.connection.object)),
            []
        )

    def test_one_page(self):
        self.connection.expect._call_.withArguments(self.ActionChecker("ListTables", {})).andReturn(_lv.ListTablesResponse(TableNames=["A", "B", "C"]))

        self.assertEqual(
            list(ListTablesIterator(self.connection.object)),
            ["A", "B", "C"]
        )

    def test_one_page_followed_by_empty_page(self):
        self.connection.expect._call_.withArguments(self.ActionChecker("ListTables", {})).andReturn(_lv.ListTablesResponse(TableNames=["A", "B", "C"], LastEvaluatedTableName="D"))
        self.connection.expect._call_.withArguments(self.ActionChecker("ListTables", {"ExclusiveStartTableName": "D"})).andReturn(_lv.ListTablesResponse(TableNames=[]))

        self.assertEqual(
            list(ListTablesIterator(self.connection.object)),
            ["A", "B", "C"]
        )

    def test_several_pages(self):
        self.connection.expect._call_.withArguments(self.ActionChecker("ListTables", {})).andReturn(_lv.ListTablesResponse(TableNames=["A", "B", "C"], LastEvaluatedTableName="D"))
        self.connection.expect._call_.withArguments(self.ActionChecker("ListTables", {"ExclusiveStartTableName": "D"})).andReturn(_lv.ListTablesResponse(TableNames=["E", "F", "G"], LastEvaluatedTableName="H"))
        self.connection.expect._call_.withArguments(self.ActionChecker("ListTables", {"ExclusiveStartTableName": "H"})).andReturn(_lv.ListTablesResponse(TableNames=["I", "J", "K"]))

        self.assertEqual(
            list(ListTablesIterator(self.connection.object)),
            ["A", "B", "C", "E", "F", "G", "I", "J", "K"]
        )


class ListTablesIteratorLocalIntegTests(_tst.LocalIntegTests):
    table_names = ["Tab{:03}".format(i) for i in range(103)]

    def setUp(self):
        super(ListTablesIteratorLocalIntegTests, self).setUp()
        for t in self.table_names:
            self.connection(
                _lv.CreateTable(t).hash_key("h", _lv.STRING).provisioned_throughput(1, 1)
            )

    def tearDown(self):
        for t in self.table_names:
            self.connection(_lv.DeleteTable(t))
        super(ListTablesIteratorLocalIntegTests, self).tearDown()

    def test(self):
        self.assertEqual(
            list(_lv.ListTablesIterator(self.connection)),
            self.table_names
        )



# coding: utf8

# Copyright 2014-2015 Vincent Jacques <vincent@vincent-jacques.net>

"""
In DynamoDB, the key attributes are typed.
Here are a few constants for those types, to be used in :class:`.CreateTable` or compared to what :class:`.DescribeTable` returns.
"""

STRING = "S"
"The 'string' attribute type"

NUMBER = "N"
"The 'number' attribute type"

BINARY = "B"
"The 'binary' attribute type"



# coding: utf8

# Copyright 2014-2015 Vincent Jacques <vincent@vincent-jacques.net>

from .batch_delete_item import BatchDeleteItem
from .batch_get_item_iterator import BatchGetItemIterator
from .batch_put_item import BatchPutItem
from .list_tables_iterator import ListTablesIterator
from .query_iterator import QueryIterator
from .scan_iterator import ScanIterator
from .wait_for_table_activation import WaitForTableActivation
from .wait_for_table_deletion import WaitForTableDeletion

# @todo Review the interfaces in this package: some accept a reference Action (Query, Scan), others don't
# Also it's weird that Actions are passed to the connection, but compounds accept the connection as an argument



# coding: utf8

# Copyright 2014-2015 Vincent Jacques <vincent@vincent-jacques.net>

"""
When given a :class:`UpdateTable`, the connection will return a :class:`UpdateTableResponse`:

.. testsetup::

    table = "LowVoltage.Tests.Doc.UpdateTable.1"
    table2 = "LowVoltage.Tests.Doc.UpdateTable.2"
    table3 = "LowVoltage.Tests.Doc.UpdateTable.3"
    table4 = "LowVoltage.Tests.Doc.UpdateTable.4"
    connection(CreateTable(table).hash_key("h", STRING).provisioned_throughput(1, 1))
    connection(
        CreateTable(table2).hash_key("h", STRING).provisioned_throughput(1, 1)
            .global_secondary_index("gsi").hash_key("hh", STRING).range_key("rr", NUMBER).provisioned_throughput(1, 1).project_all()
    )
    connection(
        CreateTable(table3).hash_key("h", STRING).provisioned_throughput(1, 1)
            .global_secondary_index("gsi").hash_key("hh", STRING).range_key("rr", NUMBER).provisioned_throughput(1, 1).project_all()
    )
    connection(CreateTable(table4).hash_key("h", STRING).provisioned_throughput(1, 1))
    WaitForTableActivation(connection, table)
    WaitForTableActivation(connection, table2)
    WaitForTableActivation(connection, table3)
    WaitForTableActivation(connection, table4)

>>> r = connection(
...   UpdateTable(table)
...     .provisioned_throughput(2, 2)
... )
>>> r
<LowVoltage.actions.update_table.UpdateTableResponse object at ...>
>>> r.table_description.table_status
u'UPDATING'

Note that you can use :func:`.WaitForTableActivation` to poll the table status until it's updated.

.. testcleanup::

    WaitForTableActivation(connection, table)
    WaitForTableActivation(connection, table2)
    WaitForTableActivation(connection, table3)
    WaitForTableActivation(connection, table4)
    connection(DeleteTable(table))
    connection(DeleteTable(table2))
    connection(DeleteTable(table3))
    connection(DeleteTable(table4))
    WaitForTableDeletion(connection, table)
    WaitForTableDeletion(connection, table2)
    WaitForTableDeletion(connection, table3)
    WaitForTableDeletion(connection, table4)
"""

import datetime

import LowVoltage as _lv
import LowVoltage.testing as _tst
from .action import Action
from .return_types import TableDescription, _is_dict


class UpdateTableResponse(object):
    """
    The `UpdateTable response <http://docs.aws.amazon.com/amazondynamodb/latest/APIReference/API_UpdateTable.html#API_UpdateTable_ResponseElements>`__.
    """

    def __init__(
        self,
        TableDescription=None,
        **dummy
    ):
        self.__table_description = TableDescription

    @property
    def table_description(self):
        """
        The description of the table you just updated.

        :type: ``None`` or :class:`.TableDescription`
        """
        if _is_dict(self.__table_description):  # pragma no branch (Defensive code)
            return TableDescription(**self.__table_description)


class UpdateTable(Action):
    """
    The `UpdateTable request <http://docs.aws.amazon.com/amazondynamodb/latest/APIReference/API_UpdateTable.html#API_UpdateTable_RequestParameters>`__.
    """

    # @todo Create and delete secondary indexes!

    def __init__(self, table_name):
        super(UpdateTable, self).__init__("UpdateTable", UpdateTableResponse)
        self.__table_name = table_name
        self.__attribute_definitions = {}
        self._read_capacity_units = None
        self._write_capacity_units = None
        self.__gsis = {}
        self.__active_index = self

    @property
    def payload(self):
        data = {"TableName": self.__table_name}
        if self.__attribute_definitions:
            data["AttributeDefinitions"] = [
                {"AttributeName": name, "AttributeType": typ}
                for name, typ in self.__attribute_definitions.iteritems()
            ]
        throughput = {}
        if self._read_capacity_units:
            throughput["ReadCapacityUnits"] = self._read_capacity_units
        if self._write_capacity_units:
            throughput["WriteCapacityUnits"] = self._write_capacity_units
        if throughput:
            data["ProvisionedThroughput"] = throughput
        if self.__gsis:
            data["GlobalSecondaryIndexUpdates"] = [i.payload for i in self.__gsis.itervalues()]
        return data

    class _Index(object):
        def __init__(self, verb, name):
            self._verb = verb
            self.__name = name
            self._hash_key = None
            self._range_key = None
            self._projection = None
            self._read_capacity_units = None
            self._write_capacity_units = None

        @property
        def payload(self):
            data = {"IndexName": self.__name}
            schema = []
            if self._hash_key:
                schema.append({"AttributeName": self._hash_key, "KeyType": "HASH"})
            if self._range_key:
                schema.append({"AttributeName": self._range_key, "KeyType": "RANGE"})
            if schema:
                data["KeySchema"] = schema
            if isinstance(self._projection, basestring):
                data["Projection"] = {"ProjectionType": self._projection}
            elif self._projection:
                data["Projection"] = {"ProjectionType": "INCLUDE", "NonKeyAttributes": self._projection}
            throughput = {}
            if self._read_capacity_units:
                throughput["ReadCapacityUnits"] = self._read_capacity_units
            if self._write_capacity_units:
                throughput["WriteCapacityUnits"] = self._write_capacity_units
            if throughput:
                data["ProvisionedThroughput"] = throughput
            return {self._verb: data}

    def hash_key(self, name, typ=None):
        """
        Set the hash key in KeySchema for the active index.
        If you provide a second argument, :meth:`attribute_definition` will be called as well.

        :raise: :exc:`.BuilderError` if called when no index is active or if the active index is not being created.

        See :meth:`create_global_secondary_index` for an example.
        """
        self.__check_active_index()
        self.__active_index._hash_key = name
        if typ is not None:
            self.attribute_definition(name, typ)
        return self

    def range_key(self, name, typ=None):
        """
        Set the range key in KeySchema for the active index.
        If you provide a second argument, :meth:`attribute_definition` will be called as well.

        :raise: :exc:`.BuilderError` if called when no index is active or if the active index is not being created.

        See :meth:`create_global_secondary_index` for an example.
        """
        self.__check_active_index()
        self.__active_index._range_key = name
        if typ is not None:
            self.attribute_definition(name, typ)
        return self

    def attribute_definition(self, name, typ):
        """
        Set the type of an attribute in AttributeDefinitions.
        Key attribute must be typed. See :mod:`.attribute_types` for constants to be passed to this method.
        """
        self.__attribute_definitions[name] = typ
        return self

    def provisioned_throughput(self, read_capacity_units, write_capacity_units):
        """
        Set the new provisioned throughput for the table or the active index.

        See :meth:`create_global_secondary_index` for an example.
        """
        self.__active_index._read_capacity_units = read_capacity_units
        self.__active_index._write_capacity_units = write_capacity_units
        return self

    def create_global_secondary_index(self, name):
        """
        Create a new GSI.
        This method sets the active index: methods like :meth:`provisioned_throughput` will apply to the index.

        >>> connection(
        ...   UpdateTable(table4)
        ...     .create_global_secondary_index("gsi")
        ...       .hash_key("hh", STRING)
        ...       .range_key("rr", NUMBER)
        ...       .project_all()
        ...       .provisioned_throughput(2, 2)
        ... )
        <LowVoltage.actions.update_table.UpdateTableResponse ...>
        """
        if name not in self.__gsis:
            self.__gsis[name] = self._Index("Create", name)
        self.__active_index = self.__gsis[name]
        return self

    def update_global_secondary_index(self, name):
        """
        Update an existing GSI.
        This method sets the active index: methods like :meth:`provisioned_throughput` will apply to the index.

        >>> connection(
        ...   UpdateTable(table2)
        ...     .update_global_secondary_index("gsi")
        ...       .provisioned_throughput(2, 2)
        ... )
        <LowVoltage.actions.update_table.UpdateTableResponse ...>
        """
        if name not in self.__gsis:
            self.__gsis[name] = self._Index("Update", name)
        self.__active_index = self.__gsis[name]
        return self

    def delete_global_secondary_index(self, name):
        """
        Mark a GSI for deletion.

        This method does not set the active index, because there is nothing to modify.

        >>> connection(
        ...   UpdateTable(table3)
        ...     .delete_global_secondary_index("gsi")
        ... )
        <LowVoltage.actions.update_table.UpdateTableResponse ...>
        """
        self.__gsis[name] = self._Index("Delete", name)
        return self

    def table(self):
        """
        Reset the active index: methods like :meth:`provisioned_throughput` will apply to the table.
        """
        self.__active_index = self
        return self

    def project_all(self):
        """
        Set ProjectionType to ALL for the active index.

        :raise: :exc:`.BuilderError` if called when no index is active or if the active index is not being created.

        See :meth:`create_global_secondary_index` for an example.
        """
        self.__check_active_index()
        self.__active_index._projection = "ALL"
        return self

    def project_keys_only(self):
        """
        Set ProjectionType to KEYS_ONLY for the active index.

        :raise: :exc:`.BuilderError` if called when no index is active or if the active index is not being created.
        """
        self.__check_active_index()
        self.__active_index._projection = "KEYS_ONLY"
        return self

    def project(self, *attrs):
        """
        Set ProjectionType to INCLUDE for the active index and add names to NonKeyAttributes.

        :raise: :exc:`.BuilderError` if called when no index is active or if the active index is not being created.
        """
        self.__check_active_index()
        if not isinstance(self.__active_index._projection, list):
            self.__active_index._projection = []
        for attr in attrs:
            if isinstance(attr, basestring):
                attr = [attr]
            self.__active_index._projection.extend(attr)
        return self

    def __check_active_index(self):
        if self.__active_index is self or self.__active_index._verb != "Create":
            raise _lv.BuilderError("No active index or active index not being created.")


class UpdateTableUnitTests(_tst.UnitTests):
    def test_name(self):
        self.assertEqual(UpdateTable("Foo").name, "UpdateTable")

    def test_no_arguments(self):
        self.assertEqual(UpdateTable("Foo").payload, {"TableName": "Foo"})

    def test_throughput(self):
        self.assertEqual(
            UpdateTable("Foo").provisioned_throughput(42, 43).payload,
            {
                "TableName": "Foo",
                "ProvisionedThroughput": {"ReadCapacityUnits": 42, "WriteCapacityUnits": 43},
            }
        )

    def test_attribute_definition(self):
        self.assertEqual(
            UpdateTable("Foo").attribute_definition("a", "B").payload,
            {
                "TableName": "Foo",
                "AttributeDefinitions": [{"AttributeName": "a", "AttributeType": "B"}],
            }
        )

    def test_create_gsi(self):
        self.assertEqual(
            UpdateTable("Foo").create_global_secondary_index("the_gsi").payload,
            {
                "TableName": "Foo",
                "GlobalSecondaryIndexUpdates": [
                    {"Create": {"IndexName": "the_gsi"}},
                ],
            }
        )

    def test_create_gsi_provisioned_throughput(self):
        self.assertEqual(
            UpdateTable("Foo").create_global_secondary_index("the_gsi").provisioned_throughput(1, 2).payload,
            {
                "TableName": "Foo",
                "GlobalSecondaryIndexUpdates": [
                    {"Create": {"IndexName": "the_gsi", "ProvisionedThroughput": {"ReadCapacityUnits": 1, "WriteCapacityUnits": 2}}},
                ],
            }
        )

    def test_create_gsi_hash_key(self):
        self.assertEqual(
            UpdateTable("Foo").create_global_secondary_index("the_gsi").hash_key("h").payload,
            {
                "TableName": "Foo",
                "GlobalSecondaryIndexUpdates": [
                    {"Create": {"IndexName": "the_gsi", "KeySchema": [{"AttributeName": "h", "KeyType": "HASH"}]}},
                ],
            }
        )

    def test_create_gsi_range_key(self):
        self.assertEqual(
            UpdateTable("Foo").create_global_secondary_index("the_gsi").range_key("r").payload,
            {
                "TableName": "Foo",
                "GlobalSecondaryIndexUpdates": [
                    {"Create": {"IndexName": "the_gsi", "KeySchema": [{"AttributeName": "r", "KeyType": "RANGE"}]}},
                ],
            }
        )

    def test_create_gsi_hash_key_with_type(self):
        self.assertEqual(
            UpdateTable("Foo").create_global_secondary_index("the_gsi").hash_key("h", "S").payload,
            {
                "TableName": "Foo",
                "GlobalSecondaryIndexUpdates": [
                    {"Create": {"IndexName": "the_gsi", "KeySchema": [{"AttributeName": "h", "KeyType": "HASH"}]}},
                ],
                "AttributeDefinitions": [{"AttributeName": "h", "AttributeType": "S"}]
            }
        )

    def test_create_gsi_range_key_with_type(self):
        self.assertEqual(
            UpdateTable("Foo").create_global_secondary_index("the_gsi").range_key("r", "N").payload,
            {
                "TableName": "Foo",
                "GlobalSecondaryIndexUpdates": [
                    {"Create": {"IndexName": "the_gsi", "KeySchema": [{"AttributeName": "r", "KeyType": "RANGE"}]}},
                ],
                "AttributeDefinitions": [{"AttributeName": "r", "AttributeType": "N"}]
            }
        )

    def test_create_gsi_project_all(self):
        self.assertEqual(
            UpdateTable("Foo").create_global_secondary_index("the_gsi").project_all().payload,
            {
                "TableName": "Foo",
                "GlobalSecondaryIndexUpdates": [
                    {"Create": {"IndexName": "the_gsi", "Projection": {"ProjectionType": "ALL"}}},
                ],
            }
        )

    def test_create_gsi_project_keys_only(self):
        self.assertEqual(
            UpdateTable("Foo").create_global_secondary_index("the_gsi").project_keys_only().payload,
            {
                "TableName": "Foo",
                "GlobalSecondaryIndexUpdates": [
                    {"Create": {"IndexName": "the_gsi", "Projection": {"ProjectionType": "KEYS_ONLY"}}},
                ],
            }
        )

    def test_create_gsi_project(self):
        self.assertEqual(
            UpdateTable("Foo").create_global_secondary_index("the_gsi").project("a", ["b", "c"]).project("d").payload,
            {
                "TableName": "Foo",
                "GlobalSecondaryIndexUpdates": [
                    {"Create": {"IndexName": "the_gsi", "Projection": {"ProjectionType": "INCLUDE", "NonKeyAttributes": ["a", "b", "c", "d"]}}},
                ],
            }
        )

    def test_update_gsi(self):
        self.assertEqual(
            UpdateTable("Foo").update_global_secondary_index("the_gsi").payload,
            {
                "TableName": "Foo",
                "GlobalSecondaryIndexUpdates": [
                    {"Update": {"IndexName": "the_gsi"}},
                ],
            }
        )

    def test_update_gsi_provisioned_throughput(self):
        self.assertEqual(
            UpdateTable("Foo").update_global_secondary_index("the_gsi").provisioned_throughput(42, 43).payload,
            {
                "TableName": "Foo",
                "GlobalSecondaryIndexUpdates": [
                    {"Update": {"IndexName": "the_gsi", "ProvisionedThroughput": {"ReadCapacityUnits": 42, "WriteCapacityUnits": 43}}},
                ],
            }
        )

    def test_delete_gsi(self):
        self.assertEqual(
            UpdateTable("Foo").delete_global_secondary_index("the_gsi").payload,
            {
                "TableName": "Foo",
                "GlobalSecondaryIndexUpdates": [
                    {"Delete": {"IndexName": "the_gsi"}},
                ],
            }
        )

    def test_back_to_update_gsi_after_back_to_table(self):
        self.assertEqual(
            UpdateTable("Foo").update_global_secondary_index("the_gsi").table().provisioned_throughput(12, 13)
                .update_global_secondary_index("the_gsi").provisioned_throughput(42, 43).payload,
            {
                "TableName": "Foo",
                "GlobalSecondaryIndexUpdates": [
                    {"Update": {"IndexName": "the_gsi", "ProvisionedThroughput": {"ReadCapacityUnits": 42, "WriteCapacityUnits": 43}}},
                ],
                "ProvisionedThroughput": {"ReadCapacityUnits": 12, "WriteCapacityUnits": 13},
            }
        )

    def test_back_to_create_gsi_after_back_to_table(self):
        self.assertEqual(
            UpdateTable("Foo").create_global_secondary_index("the_gsi").table().provisioned_throughput(12, 13)
                .create_global_secondary_index("the_gsi").provisioned_throughput(42, 43).payload,
            {
                "TableName": "Foo",
                "GlobalSecondaryIndexUpdates": [
                    {"Create": {"IndexName": "the_gsi", "ProvisionedThroughput": {"ReadCapacityUnits": 42, "WriteCapacityUnits": 43}}},
                ],
                "ProvisionedThroughput": {"ReadCapacityUnits": 12, "WriteCapacityUnits": 13},
            }
        )

    def test_back_to_update_gsi_after_back_to_table_after_create_gsi(self):
        self.assertEqual(
            UpdateTable("Foo").create_global_secondary_index("the_gsi").table().provisioned_throughput(12, 13)
                .update_global_secondary_index("the_gsi").provisioned_throughput(42, 43).payload,
            {
                "TableName": "Foo",
                "GlobalSecondaryIndexUpdates": [
                    {"Create": {"IndexName": "the_gsi", "ProvisionedThroughput": {"ReadCapacityUnits": 42, "WriteCapacityUnits": 43}}},
                ],
                "ProvisionedThroughput": {"ReadCapacityUnits": 12, "WriteCapacityUnits": 13},
            }
        )

    def test_hash_key_without_active_index(self):
        with self.assertRaises(_lv.BuilderError) as catcher:
            UpdateTable("Foo").hash_key("h")
        self.assertEqual(catcher.exception.args, ("No active index or active index not being created.",))

    def test_range_key_without_active_index(self):
        with self.assertRaises(_lv.BuilderError) as catcher:
            UpdateTable("Foo").range_key("r")
        self.assertEqual(catcher.exception.args, ("No active index or active index not being created.",))

    def test_project_all_without_active_index(self):
        with self.assertRaises(_lv.BuilderError) as catcher:
            UpdateTable("Foo").project_all()
        self.assertEqual(catcher.exception.args, ("No active index or active index not being created.",))

    def test_project_without_active_index(self):
        with self.assertRaises(_lv.BuilderError) as catcher:
            UpdateTable("Foo").project("a")
        self.assertEqual(catcher.exception.args, ("No active index or active index not being created.",))

    def test_project_keys_only_without_active_index(self):
        with self.assertRaises(_lv.BuilderError) as catcher:
            UpdateTable("Foo").project_keys_only()
        self.assertEqual(catcher.exception.args, ("No active index or active index not being created.",))

    def test_hash_key_with_updating_active_index(self):
        with self.assertRaises(_lv.BuilderError) as catcher:
            UpdateTable("Foo").update_global_secondary_index("gsi").hash_key("h")
        self.assertEqual(catcher.exception.args, ("No active index or active index not being created.",))

    def test_range_key_with_updating_active_index(self):
        with self.assertRaises(_lv.BuilderError) as catcher:
            UpdateTable("Foo").update_global_secondary_index("gsi").range_key("r")
        self.assertEqual(catcher.exception.args, ("No active index or active index not being created.",))

    def test_project_all_with_updating_active_index(self):
        with self.assertRaises(_lv.BuilderError) as catcher:
            UpdateTable("Foo").update_global_secondary_index("gsi").project_all()
        self.assertEqual(catcher.exception.args, ("No active index or active index not being created.",))

    def test_project_with_updating_active_index(self):
        with self.assertRaises(_lv.BuilderError) as catcher:
            UpdateTable("Foo").update_global_secondary_index("gsi").project("a")
        self.assertEqual(catcher.exception.args, ("No active index or active index not being created.",))

    def test_project_keys_only_with_updating_active_index(self):
        with self.assertRaises(_lv.BuilderError) as catcher:
            UpdateTable("Foo").update_global_secondary_index("gsi").project_keys_only()
        self.assertEqual(catcher.exception.args, ("No active index or active index not being created.",))


class UpdateTableLocalIntegTests(_tst.LocalIntegTests):
    def setUp(self):
        super(UpdateTableLocalIntegTests, self).setUp()
        self.connection(
            _lv.CreateTable("Aaa").hash_key("h", _lv.STRING).provisioned_throughput(1, 2)
                .global_secondary_index("the_gsi")
                .hash_key("hh", _lv.STRING)
                .project_all()
                .provisioned_throughput(3, 4)
        )

    def tearDown(self):
        self.connection(_lv.DeleteTable("Aaa"))
        super(UpdateTableLocalIntegTests, self).tearDown()

    def test_provisioned_throughput(self):
        r = self.connection(
            _lv.UpdateTable("Aaa").provisioned_throughput(2, 4)
        )

        self.assertEqual(r.table_description.provisioned_throughput.read_capacity_units, 2)
        self.assertEqual(r.table_description.provisioned_throughput.write_capacity_units, 4)

    def test_update_gsi_provisioned_throughput(self):
        r = self.connection(
            _lv.UpdateTable("Aaa").update_global_secondary_index("the_gsi").provisioned_throughput(6, 8)
        )

        self.assertEqual(r.table_description.global_secondary_indexes[0].provisioned_throughput.read_capacity_units, 6)
        self.assertEqual(r.table_description.global_secondary_indexes[0].provisioned_throughput.write_capacity_units, 8)

    def test_delete_and_create_gsi(self):
        r = self.connection(_lv.UpdateTable("Aaa").delete_global_secondary_index("the_gsi"))
        self.assertEqual(r.table_description.global_secondary_indexes[0].index_status, "DELETING")

        _lv.WaitForTableActivation(self.connection, "Aaa")

        r = self.connection(_lv.DescribeTable("Aaa"))
        self.assertEqual(r.table.global_secondary_indexes, None)
        self.assertEqual(len(r.table.attribute_definitions), 1)  # The previous definition of attribute "hh" has disapeared.

        r = self.connection(
            _lv.UpdateTable("Aaa")
                .create_global_secondary_index("new_gsi")
                    .provisioned_throughput(1, 2)
                    .hash_key("nh", _lv.NUMBER)
                    .project_all()
        )

        self.assertEqual(r.table_description.global_secondary_indexes[0].provisioned_throughput.read_capacity_units, 1)
        self.assertEqual(r.table_description.global_secondary_indexes[0].provisioned_throughput.write_capacity_units, 2)
        self.assertEqual(r.table_description.global_secondary_indexes[0].key_schema[0].attribute_name, "nh")
        self.assertEqual(r.table_description.global_secondary_indexes[0].key_schema[0].key_type, "HASH")
        self.assertEqual(len(r.table_description.attribute_definitions), 2)



# coding: utf8

# Copyright 2014-2015 Vincent Jacques <vincent@vincent-jacques.net>

"""
When given a :class:`BatchGetItem`, the connection will return a :class:`BatchGetItemResponse`:

>>> connection(BatchGetItem().table(table).keys({"h": 0}, {"h": 1}))
<LowVoltage.actions.batch_get_item.BatchGetItemResponse ...>

Responses are accessed like this:

>>> connection(
...   BatchGetItem().table(table).keys({"h": 0})
... ).responses[table]
[{u'h': 0, u'gr': 0, u'gh': 0}]

Note that responses are in an undefined order.
"""

import LowVoltage as _lv
import LowVoltage.testing as _tst
from .action import Action
from .conversion import _convert_dict_to_db, _convert_db_to_dict
from .next_gen_mixins import proxy
from .next_gen_mixins import (
    ConsistentRead,
    ExpressionAttributeNames,
    ProjectionExpression,
    ReturnConsumedCapacity,
)
from .return_types import ConsumedCapacity, _is_dict, _is_list_of_dict


class BatchGetItemResponse(object):
    """
    The `BatchGetItem response <http://docs.aws.amazon.com/amazondynamodb/latest/APIReference/API_BatchGetItem.html#API_BatchGetItem_ResponseElements>`__
    """

    def __init__(
        self,
        ConsumedCapacity=None,
        Responses=None,
        UnprocessedKeys=None,
        **dummy
    ):
        self.__consumed_capacity = ConsumedCapacity
        self.__responses = Responses
        self.__unprocessed_keys = UnprocessedKeys

    @property
    def consumed_capacity(self):
        """
        The capacity consumed by the request. If you used :meth:`~BatchGetItem.return_consumed_capacity_total`.

        :type: ``None`` or list of :class:`.ConsumedCapacity`
        """
        if _is_list_of_dict(self.__consumed_capacity):  # pragma no branch (Defensive code)
            return [ConsumedCapacity(**c) for c in self.__consumed_capacity]

    @property
    def responses(self):
        """
        The items you just got.

        :type: ``None`` or dict of string (table name) to list of dict
        """
        if _is_dict(self.__responses):  # pragma no branch (Defensive code)
            return {t: [_convert_db_to_dict(v) for v in vs] for t, vs in self.__responses.iteritems()}

    @property
    def unprocessed_keys(self):
        """
        Keys that were not processed during this request. If not None, you should give this back to the constructor of a subsequent :class:`BatchGetItem`.

        :type: ``None`` or exactly as returned by DynamoDB
        """
        return self.__unprocessed_keys


class BatchGetItem(Action):
    """
    The `BatchGetItem request <http://docs.aws.amazon.com/amazondynamodb/latest/APIReference/API_BatchGetItem.html#API_BatchGetItem_RequestParameters>`__
    """

    def __init__(self, previous_unprocessed_keys=None):
        super(BatchGetItem, self).__init__("BatchGetItem", BatchGetItemResponse)
        self.__previous_unprocessed_keys = previous_unprocessed_keys
        self.__tables = {}
        self.__active_table = None
        self.__return_consumed_capacity = ReturnConsumedCapacity(self)

    @property
    def payload(self):
        data = {}
        if self.__previous_unprocessed_keys:
            data["RequestItems"] = self.__previous_unprocessed_keys
        if self.__tables:
            data["RequestItems"] = {n: t.payload for n, t in self.__tables.iteritems()}
        data.update(self.__return_consumed_capacity.payload)
        return data

    class _Table:
        def __init__(self, action):
            self.keys = []
            self.consistent_read = ConsistentRead(action)
            self.expression_attribute_names = ExpressionAttributeNames(action)
            self.projection_expression = ProjectionExpression(action)

        @property
        def payload(self):
            data = {}
            if self.keys:
                data["Keys"] = [_convert_dict_to_db(k) for k in self.keys]
            data.update(self.consistent_read.payload)
            data.update(self.expression_attribute_names.payload)
            data.update(self.projection_expression.payload)
            return data

    def table(self, name):
        """
        Set the active table. Calls to methods like :meth:`keys` or :meth:`consistent_read_true` will apply to this table.
        """
        if name not in self.__tables:
            self.__tables[name] = self._Table(self)
        self.__active_table = self.__tables[name]
        return self

    # @todo Add unit tests about calling keys (and others) without an active table. Same in other multi-table actions. Same in CreateTable for active index.

    def keys(self, *keys):
        """
        Add keys to get from the active table.
        This method accepts a variable number of keys or iterables.

        :raise: :exc:`.BuilderError` if called when no table is active.

        >>> connection(
        ...   BatchGetItem()
        ...     .table(table)
        ...     .keys({"h": 1})
        ...     .keys({"h": 2}, {"h": 3})
        ...     .keys([{"h": 4}, {"h": 5}])
        ...     .keys({"h": h} for h in range(6, 10))
        ... )
        <LowVoltage.actions.batch_get_item.BatchGetItemResponse ...>
        """
        self.__check_active_table()
        for key in keys:
            if isinstance(key, dict):
                key = [key]
            self.__active_table.keys.extend(key)
        return self

    @proxy
    def consistent_read_true(self):
        """
        :raise: :exc:`.BuilderError` if called when no table is active.

        >>> c = connection(
        ...   BatchGetItem()
        ...     .table(table).keys({"h": 0})
        ...     .consistent_read_true()
        ...     .return_consumed_capacity_total()
        ... ).consumed_capacity
        >>> c[0].table_name
        u'LowVoltage.Tests.Doc.1'
        >>> c[0].capacity_units
        1.0
        """
        self.__check_active_table()
        return self.__active_table.consistent_read.true()

    @proxy
    def consistent_read_false(self):
        """
        :raise: :exc:`.BuilderError` if called when no table is active.

        >>> c = connection(
        ...   BatchGetItem()
        ...     .table(table).keys({"h": 0})
        ...     .consistent_read_false()
        ...     .return_consumed_capacity_total()
        ... ).consumed_capacity
        >>> c[0].table_name
        u'LowVoltage.Tests.Doc.1'
        >>> c[0].capacity_units
        0.5
        """
        self.__check_active_table()
        return self.__active_table.consistent_read.false()

    @proxy
    def project(self, *names):
        """
        :raise: :exc:`.BuilderError` if called when no table is active.

        >>> connection(
        ...   BatchGetItem()
        ...     .table(table)
        ...     .keys({"h": 0})
        ...     .project("h", "gr")
        ... ).responses[table]
        [{u'h': 0, u'gr': 0}]
        """
        self.__check_active_table()
        self.__active_table.projection_expression.add(*names)
        return self

    @proxy
    def expression_attribute_name(self, name, path):
        """
        :raise: :exc:`.BuilderError` if called when no table is active.

        >>> connection(
        ...   BatchGetItem()
        ...     .table(table)
        ...     .keys({"h": 0})
        ...     .expression_attribute_name("syn", "h")
        ...     .project("#syn")
        ... ).responses[table]
        [{u'h': 0}]
        """
        self.__check_active_table()
        self.__active_table.expression_attribute_names.add(name, path)
        return self

    @proxy
    def return_consumed_capacity_total(self):
        """
        >>> c = connection(
        ...   BatchGetItem()
        ...     .table(table).keys({"h": 0}, {"h": 1}, {"h": 2})
        ...     .table(table2).keys({"h": 0, "r1": 0}, {"h": 1, "r1": 0})
        ...     .return_consumed_capacity_total()
        ... ).consumed_capacity
        >>> c[0].table_name
        u'LowVoltage.Tests.Doc.1'
        >>> c[0].capacity_units
        1.5
        >>> c[1].table_name
        u'LowVoltage.Tests.Doc.2'
        >>> c[1].capacity_units
        1.0
        """
        return self.__return_consumed_capacity.total()

    @proxy
    def return_consumed_capacity_none(self):
        """
        >>> print connection(
        ...   BatchGetItem()
        ...     .table(table)
        ...     .keys({"h": 0}, {"h": 1}, {"h": 2})
        ...     .return_consumed_capacity_none()
        ... ).consumed_capacity
        None
        """
        return self.__return_consumed_capacity.none()

    def __check_active_table(self):
        if self.__active_table is None:
            raise _lv.BuilderError("No active table.")


class BatchGetItemUnitTests(_tst.UnitTests):
    def test_name(self):
        self.assertEqual(BatchGetItem().name, "BatchGetItem")

    def test_empty(self):
        self.assertEqual(
            BatchGetItem().payload,
            {}
        )

    def test_return_consumed_capacity_none(self):
        self.assertEqual(
            BatchGetItem().return_consumed_capacity_none().payload,
            {
                "ReturnConsumedCapacity": "NONE",
            }
        )

    def test_return_consumed_capacity_total(self):
        self.assertEqual(
            BatchGetItem().return_consumed_capacity_total().payload,
            {
                "ReturnConsumedCapacity": "TOTAL",
            }
        )

    def test_keys(self):
        self.assertEqual(
            BatchGetItem().table("Table2").keys({"hash": u"h21"}).table("Table1").keys({"hash": u"h11"}, {"hash": u"h12"}).table("Table2").keys([{"hash": u"h22"}, {"hash": u"h23"}]).payload,
            {
                "RequestItems": {
                    "Table1": {
                        "Keys": [
                            {"hash": {"S": "h11"}},
                            {"hash": {"S": "h12"}},
                        ]
                    },
                    "Table2": {
                        "Keys": [
                            {"hash": {"S": "h21"}},
                            {"hash": {"S": "h22"}},
                            {"hash": {"S": "h23"}},
                        ]
                    },
                }
            }
        )

    def test_consistent_read(self):
        self.assertEqual(
            BatchGetItem().table("Table1").consistent_read_true().table("Table2").consistent_read_false().payload,
            {
                "RequestItems": {
                    "Table1": {
                        "ConsistentRead": True,
                    },
                    "Table2": {
                        "ConsistentRead": False,
                    },
                }
            }
        )

    def test_project(self):
        self.assertEqual(
            BatchGetItem().table("Table1").project("a").payload,
            {
                "RequestItems": {
                    "Table1": {
                        "ProjectionExpression": "a",
                    },
                }
            }
        )

    def test_expression_attribute_name(self):
        self.assertEqual(
            BatchGetItem().table("Table1").expression_attribute_name("a", "p").payload,
            {
                "RequestItems": {
                    "Table1": {
                        "ExpressionAttributeNames": {"#a": "p"},
                    },
                }
            }
        )

    def test_keys_without_active_table(self):
        with self.assertRaises(_lv.BuilderError) as catcher:
            BatchGetItem().keys({"h": 0})
        self.assertEqual(catcher.exception.args, ("No active table.",))

    def test_consistent_read_true_without_active_table(self):
        with self.assertRaises(_lv.BuilderError) as catcher:
            BatchGetItem().consistent_read_true()
        self.assertEqual(catcher.exception.args, ("No active table.",))

    def test_consistent_read_false_without_active_table(self):
        with self.assertRaises(_lv.BuilderError) as catcher:
            BatchGetItem().consistent_read_false()
        self.assertEqual(catcher.exception.args, ("No active table.",))

    def test_project_without_active_table(self):
        with self.assertRaises(_lv.BuilderError) as catcher:
            BatchGetItem().project("a")
        self.assertEqual(catcher.exception.args, ("No active table.",))

    def test_expression_attribute_name_without_active_table(self):
        with self.assertRaises(_lv.BuilderError) as catcher:
            BatchGetItem().expression_attribute_name("a", "b")
        self.assertEqual(catcher.exception.args, ("No active table.",))


class BatchGetItemLocalIntegTests(_tst.LocalIntegTestsWithTableH):
    def test_simple_batch_get(self):
        self.connection(_lv.BatchWriteItem().table("Aaa").put(
            {"h": u"1", "a": "xxx"},
            {"h": u"2", "a": "yyy"},
            {"h": u"3", "a": "zzz"},
        ))

        r = self.connection(_lv.BatchGetItem().table("Aaa").keys({"h": u"1"}, {"h": u"2"}, {"h": u"3"}))

        self.assertEqual(r.responses.keys(), ["Aaa"])
        self.assertEqual(
            sorted(r.responses["Aaa"], key=lambda i: i["h"]),
            [{"h": u"1", "a": "xxx"}, {"h": u"2", "a": "yyy"}, {"h": u"3", "a": "zzz"}]
        )

    def test_batch_get_with_projections(self):
        self.connection(_lv.BatchWriteItem().table("Aaa").put(
            {"h": u"1", "a": "a1", "b": "b1", "c": "c1"},
            {"h": u"2", "a": "a2", "b": "b2", "c": "c2"},
            {"h": u"3", "a": "a3", "b": "b3", "c": "c3"},
        ))

        r = self.connection(
            _lv.BatchGetItem().table("Aaa").keys({"h": u"1"}, {"h": u"2"}, {"h": u"3"}).expression_attribute_name("p", "b").project("h").project("a", ["#p"])
        )
        self.assertEqual(
            sorted(r.responses["Aaa"], key=lambda i: i["h"]),
            [{"h": u"1", "a": "a1", "b": "b1"}, {"h": u"2", "a": "a2", "b": "b2"}, {"h": u"3", "a": "a3", "b": "b3"}]
        )

    def test_get_unexisting_keys(self):
        self.connection(_lv.BatchWriteItem().table("Aaa").put(
            {"h": u"1", "a": "xxx"},
            {"h": u"2", "a": "yyy"},
        ))

        r = self.connection(_lv.BatchGetItem().table("Aaa").keys({"h": u"1"}, {"h": u"2"}, {"h": u"3"}))

        self.assertEqual(
            sorted(r.responses["Aaa"], key=lambda i: i["h"]),
            [{"h": u"1", "a": "xxx"}, {"h": u"2", "a": "yyy"}]
        )
        self.assertEqual(r.unprocessed_keys, {})

    def test_get_without_unprocessed_keys(self):
        _lv.BatchPutItem(self.connection, "Aaa", [{"h": unicode(i)} for i in range(100)])

        r = self.connection(_lv.BatchGetItem().table("Aaa").keys({"h": unicode(i)} for i in range(100)))

        self.assertEqual(r.unprocessed_keys, {})
        self.assertEqual(len(r.responses["Aaa"]), 100)

    def test_get_with_unprocessed_keys(self):
        _lv.BatchPutItem(self.connection, "Aaa", [{"h": unicode(i), "xs": "x" * 300000} for i in range(100)])  # 300kB items ensure a single BatchGetItem will return at most 55 items

        r1 = self.connection(_lv.BatchGetItem().table("Aaa").keys({"h": unicode(i)} for i in range(100)))

        self.assertEqual(len(r1.unprocessed_keys["Aaa"]["Keys"]), 45)
        self.assertEqual(len(r1.responses["Aaa"]), 55)


class BatchGetItemConnectedIntegTests(_tst.ConnectedIntegTestsWithTable):
    def setUp(self):
        super(BatchGetItemConnectedIntegTests, self).setUp()
        self.connection(_lv.PutItem(self.table, self.item))

    def tearDown(self):
        self.connection(_lv.DeleteItem(self.table, self.tab_key))
        super(BatchGetItemConnectedIntegTests, self).tearDown()

    def test_return_consumed_capacity_total(self):
        r = self.connection(_lv.BatchGetItem().table(self.table).keys(self.tab_key).return_consumed_capacity_total())

        self.assertEqual(r.consumed_capacity[0].capacity_units, 0.5)
        self.assertEqual(r.consumed_capacity[0].global_secondary_indexes, None)
        self.assertEqual(r.consumed_capacity[0].local_secondary_indexes, None)
        self.assertEqual(r.consumed_capacity[0].table, None)
        self.assertEqual(r.consumed_capacity[0].table_name, self.table)



# coding: utf8

# Copyright 2014-2015 Vincent Jacques <vincent@vincent-jacques.net>

from ..conversion import ConversionUnitTests
from ..expressions import ConditionExpressionUnitTests

from ..batch_get_item import BatchGetItemUnitTests
from ..batch_write_item import BatchWriteItemUnitTests
from ..create_table import CreateTableUnitTests
from ..delete_item import DeleteItemUnitTests
from ..delete_table import DeleteTableUnitTests
from ..describe_table import DescribeTableUnitTests
from ..get_item import GetItemUnitTests
from ..list_tables import ListTablesUnitTests
from ..put_item import PutItemUnitTests
from ..query import QueryUnitTests
from ..scan import ScanUnitTests
from ..update_item import UpdateItemUnitTests
from ..update_table import UpdateTableUnitTests



# coding: utf8

# Copyright 2014-2015 Vincent Jacques <vincent@vincent-jacques.net>

"""
When given a :class:`Scan`, the connection will return a :class:`ScanResponse`:

>>> connection(Scan(table))
<LowVoltage.actions.scan.ScanResponse ...>

Items are accessed like this:

>>> connection(Scan(table)).items
[{u'h': 7, u'gr': 0, u'gh': 0}, {u'h': 8, u'gr': 0, u'gh': 0}, {u'h': 3, u'gr': 0, u'gh': 0}, {u'h': 2, u'gr': 0, u'gh': 0}, {u'h': 9, u'gr': 0, u'gh': 0}, {u'h': 4, u'gr': 0, u'gh': 0}, {u'h': 6, u'gr': 0, u'gh': 0}, {u'h': 1, u'gr': 0, u'gh': 0}, {u'h': 0, u'gr': 0, u'gh': 0}, {u'h': 5, u'gr': 0, u'gh': 0}]

Note that items are returned in an undefined order.
"""

import LowVoltage as _lv
import LowVoltage.testing as _tst
from .action import Action
from .conversion import _convert_db_to_dict
from .next_gen_mixins import proxy
from .next_gen_mixins import (
    ExclusiveStartKey,
    ExpressionAttributeNames,
    ExpressionAttributeValues,
    FilterExpression,
    Limit,
    ProjectionExpression,
    ReturnConsumedCapacity,
    ScalarValue,
    Select,
)
from .return_types import ConsumedCapacity, _is_dict, _is_int, _is_list_of_dict


class ScanResponse(object):
    """
    The `Scan response <http://docs.aws.amazon.com/amazondynamodb/latest/APIReference/API_Scan.html#API_Scan_ResponseElements>`__.
    """

    def __init__(
        self,
        ConsumedCapacity=None,
        Count=None,
        Items=None,
        LastEvaluatedKey=None,
        ScannedCount=None,
        **dummy
    ):
        self.__consumed_capacity = ConsumedCapacity
        self.__count = Count
        self.__items = Items
        self.__last_evaluated_key = LastEvaluatedKey
        self.__scanned_count = ScannedCount

    @property
    def consumed_capacity(self):
        """
        The capacity consumed by the request. If you used :meth:`~Scan.return_consumed_capacity_total`.

        :type: ``None`` or :class:`.ConsumedCapacity`
        """
        if _is_dict(self.__consumed_capacity):  # pragma no branch (Defensive code)
            return ConsumedCapacity(**self.__consumed_capacity)

    @property
    def count(self):
        """
        The number of items matching the scan.

        :type: ``None`` or long
        """
        if _is_int(self.__count):  # pragma no branch (Defensive code)
            return long(self.__count)

    @property
    def items(self):
        """
        The items matching the scan. Unless you used :meth:`.Scan.select_count`.

        :type: ``None`` or list of dict
        """
        if _is_list_of_dict(self.__items):  # pragma no branch (Defensive code)
            return [_convert_db_to_dict(i) for i in self.__items]

    @property
    def last_evaluated_key(self):
        """
        The key of the last item evaluated by the scan. If not None, it should be given to :meth:`~Scan.exclusive_start_key` is a subsequent :class:`Scan`.

        :type: ``None`` or dict
        """
        if _is_dict(self.__last_evaluated_key):  # pragma no branch (Defensive code)
            return _convert_db_to_dict(self.__last_evaluated_key)

    @property
    def scanned_count(self):
        """
        The number of item scanned during the scan. This can be different from :attr:`count` when using :meth:`~Scan.filter_expression`.

        :type: ``None`` or long
        """
        if _is_int(self.__scanned_count):  # pragma no branch (Defensive code)
            return long(self.__scanned_count)


class Scan(Action):
    """
    The `Scan request <http://docs.aws.amazon.com/amazondynamodb/latest/APIReference/API_Scan.html#API_Scan_RequestParameters>`__.
    """

    def __init__(self, table_name):
        super(Scan, self).__init__("Scan", ScanResponse)
        self.__table_name = table_name
        self.__exclusive_start_key = ExclusiveStartKey(self)
        self.__expression_attribute_names = ExpressionAttributeNames(self)
        self.__expression_attribute_values = ExpressionAttributeValues(self)
        self.__filter_expression = FilterExpression(self)
        self.__limit = Limit(self)
        self.__projection_expression = ProjectionExpression(self)
        self.__return_consumed_capacity = ReturnConsumedCapacity(self)
        self.__segment = ScalarValue("Segment")(self)
        self.__select = Select(self)
        self.__total_segments = ScalarValue("TotalSegments")(self)

    @property
    def payload(self):
        data = {"TableName": self.__table_name}
        data.update(self.__exclusive_start_key.payload)
        data.update(self.__expression_attribute_names.payload)
        data.update(self.__expression_attribute_values.payload)
        data.update(self.__filter_expression.payload)
        data.update(self.__limit.payload)
        data.update(self.__projection_expression.payload)
        data.update(self.__return_consumed_capacity.payload)
        data.update(self.__segment.payload)
        data.update(self.__select.payload)
        data.update(self.__total_segments.payload)
        return data

    def segment(self, segment, total_segments):
        """
        Set Segment and TotalSegments for a `parallel scan <http://docs.aws.amazon.com/amazondynamodb/latest/developerguide/QueryAndScan.html#QueryAndScanParallelScan>`__.

        Items will be partitioned in ``total_segments`` segments of approximately the same size,
        ans only the items of the ``segment``-th segment will be returned in this request.

        See also :meth:`.ScanIterator.parallelize`.

        >>> connection(
        ...   Scan(table)
        ...     .project("h")
        ...     .segment(0, 2)
        ... ).items
        [{u'h': 7}, {u'h': 8}, {u'h': 3}, {u'h': 2}, {u'h': 9}, {u'h': 4}]
        >>> connection(
        ...   Scan(table)
        ...     .project("h")
        ...     .segment(1, 2)
        ... ).items
        [{u'h': 6}, {u'h': 1}, {u'h': 0}, {u'h': 5}]
        """
        self.__segment.set(segment)
        return self.__total_segments.set(total_segments)

    @proxy("Scan")
    def exclusive_start_key(self, key):
        """
        >>> r = connection(
        ...   Scan(table)
        ...     .project("h")
        ...     .limit(5)
        ... )
        >>> r.items
        [{u'h': 7}, {u'h': 8}, {u'h': 3}, {u'h': 2}, {u'h': 9}]
        >>> r.last_evaluated_key
        {u'h': 9}
        >>> connection(
        ...   Scan(table)
        ...     .project("h")
        ...     .exclusive_start_key({"h": 9})
        ... ).items
        [{u'h': 4}, {u'h': 6}, {u'h': 1}, {u'h': 0}, {u'h': 5}]
        """
        return self.__exclusive_start_key.set(key)

    @proxy
    def limit(self, limit):
        """
        See :meth:`exclusive_start_key` for an example.
        """
        return self.__limit.set(limit)

    @proxy
    def select_count(self):
        """
        >>> r = connection(Scan(table).select_count())
        >>> r.count
        10L
        >>> print r.items
        None
        """
        return self.__select.count()

    @proxy
    def select_all_attributes(self):
        """
        >>> connection(Scan(table).select_all_attributes()).items
        [{u'h': 7, u'gr': 0, u'gh': 0}, {u'h': 8, u'gr': 0, u'gh': 0}, {u'h': 3, u'gr': 0, u'gh': 0}, {u'h': 2, u'gr': 0, u'gh': 0}, {u'h': 9, u'gr': 0, u'gh': 0}, {u'h': 4, u'gr': 0, u'gh': 0}, {u'h': 6, u'gr': 0, u'gh': 0}, {u'h': 1, u'gr': 0, u'gh': 0}, {u'h': 0, u'gr': 0, u'gh': 0}, {u'h': 5, u'gr': 0, u'gh': 0}]
        """
        return self.__select.all_attributes()

    @proxy
    def project(self, *names):
        """
        >>> connection(Scan(table).project("h")).items
        [{u'h': 7}, {u'h': 8}, {u'h': 3}, {u'h': 2}, {u'h': 9}, {u'h': 4}, {u'h': 6}, {u'h': 1}, {u'h': 0}, {u'h': 5}]
        """
        return self.__projection_expression.add(*names)

    @proxy
    def filter_expression(self, expression):
        """
        >>> connection(
        ...   Scan(table)
        ...     .filter_expression("#syn=:val")
        ...     .expression_attribute_name("syn", "a")
        ...     .expression_attribute_value("val", 42)
        ... ).items
        []
        """
        return self.__filter_expression.set(expression)

    @proxy
    def expression_attribute_name(self, synonym, name):
        """
        See :meth:`filter_expression` for an example.
        """
        return self.__expression_attribute_names.add(synonym, name)

    @proxy
    def expression_attribute_value(self, name, value):
        """
        See :meth:`filter_expression` for an example.
        """
        return self.__expression_attribute_values.add(name, value)

    @proxy
    def return_consumed_capacity_total(self):
        """
        >>> connection(
        ...   Scan(table)
        ...     .return_consumed_capacity_total()
        ... ).consumed_capacity.capacity_units
        0.5
        """
        return self.__return_consumed_capacity.total()

    @proxy
    def return_consumed_capacity_none(self):
        """
        >>> print connection(
        ...   Scan(table)
        ...     .return_consumed_capacity_none()
        ... ).consumed_capacity
        None
        """
        return self.__return_consumed_capacity.none()


class ScanUnitTests(_tst.UnitTests):
    def test_name(self):
        self.assertEqual(Scan("Aaa").name, "Scan")

    def test_table_name(self):
        self.assertEqual(Scan("Aaa").payload, {"TableName": "Aaa"})

    def test_segment(self):
        self.assertEqual(Scan("Aaa").segment(0, 2).payload, {"TableName": "Aaa", "Segment": 0, "TotalSegments": 2})

    def test_exclusive_start_key(self):
        self.assertEqual(Scan("Aaa").exclusive_start_key({"h": u"v"}).payload, {"TableName": "Aaa", "ExclusiveStartKey": {"h": {"S": "v"}}})

    def test_limit(self):
        self.assertEqual(Scan("Aaa").limit(4).payload, {"TableName": "Aaa", "Limit": 4})

    def test_select_all_attributes(self):
        self.assertEqual(Scan("Aaa").select_all_attributes().payload, {"TableName": "Aaa", "Select": "ALL_ATTRIBUTES"})

    def test_select_count(self):
        self.assertEqual(Scan("Aaa").select_count().payload, {"TableName": "Aaa", "Select": "COUNT"})

    def test_expression_attribute_name(self):
        self.assertEqual(Scan("Aaa").expression_attribute_name("n", "p").payload, {"TableName": "Aaa", "ExpressionAttributeNames": {"#n": "p"}})

    def test_expression_attribute_value(self):
        self.assertEqual(Scan("Aaa").expression_attribute_value("n", u"p").payload, {"TableName": "Aaa", "ExpressionAttributeValues": {":n": {"S": "p"}}})

    def test_project(self):
        self.assertEqual(Scan("Aaa").project("a").payload, {"TableName": "Aaa", "ProjectionExpression": "a"})

    def test_return_consumed_capacity_total(self):
        self.assertEqual(Scan("Aaa").return_consumed_capacity_total().payload, {"TableName": "Aaa", "ReturnConsumedCapacity": "TOTAL"})

    def test_return_consumed_capacity_none(self):
        self.assertEqual(Scan("Aaa").return_consumed_capacity_none().payload, {"TableName": "Aaa", "ReturnConsumedCapacity": "NONE"})

    def test_filter_expression(self):
        self.assertEqual(Scan("Aaa").filter_expression("a=b").payload, {"TableName": "Aaa", "FilterExpression": "a=b"})


class ScanLocalIntegTests(_tst.LocalIntegTestsWithTableH):
    def setUp(self):
        super(ScanLocalIntegTests, self).setUp()
        self.connection(_lv.BatchWriteItem().table("Aaa").put(
            {"h": u"0", "v": 0},
            {"h": u"1", "v": 1},
            {"h": u"2", "v": 2},
            {"h": u"3", "v": 3},
        ))

    def test_simple_scan(self):
        r = self.connection(
            _lv.Scan("Aaa")
        )

        self.assertEqual(r.count, 4)
        items = sorted((r.items[i] for i in range(4)), key=lambda i: i["h"])
        self.assertEqual(items, [{"h": u"0", "v": 0}, {"h": u"1", "v": 1}, {"h": u"2", "v": 2}, {"h": u"3", "v": 3}])
        self.assertEqual(r.last_evaluated_key, None)
        self.assertEqual(r.scanned_count, 4)

    def test_paginated_segmented_scan(self):
        # If this test fails randomly, change it to assert on the sum and union of the results
        r01 = self.connection(
            _lv.Scan("Aaa").segment(0, 2).limit(1)
        )
        r02 = self.connection(
            _lv.Scan("Aaa").segment(0, 2).exclusive_start_key({"h": u"1"})
        )
        r11 = self.connection(
            _lv.Scan("Aaa").segment(1, 2).limit(1)
        )
        r12 = self.connection(
            _lv.Scan("Aaa").segment(1, 2).exclusive_start_key({"h": u"0"})
        )

        self.assertEqual(r01.count, 1)
        self.assertEqual(r01.items[0], {"h": u"1", "v": 1})
        self.assertEqual(r01.last_evaluated_key, {"h": u"1"})
        self.assertEqual(r01.scanned_count, 1)

        self.assertEqual(r02.count, 1)
        self.assertEqual(r02.items[0], {"h": u"3", "v": 3})
        self.assertEqual(r02.last_evaluated_key, None)
        self.assertEqual(r02.scanned_count, 1)

        self.assertEqual(r11.count, 1)
        self.assertEqual(r11.items[0], {"h": u"0", "v": 0})
        self.assertEqual(r11.last_evaluated_key, {"h": u"0"})
        self.assertEqual(r11.scanned_count, 1)

        self.assertEqual(r12.count, 1)
        self.assertEqual(r12.items[0], {"h": u"2", "v": 2})
        self.assertEqual(r12.last_evaluated_key, None)
        self.assertEqual(r12.scanned_count, 1)

    def test_filtered_scan(self):
        r = self.connection(
            _lv.Scan("Aaa").filter_expression("v>:v").expression_attribute_value("v", 1).project("h")
        )

        self.assertEqual(r.count, 2)
        self.assertEqual(r.items[0], {"h": u"3"})
        self.assertEqual(r.items[1], {"h": u"2"})
        self.assertEqual(r.last_evaluated_key, None)
        self.assertEqual(r.scanned_count, 4)


class ScanConnectedIntegTests(_tst.ConnectedIntegTestsWithTable):
    def setUp(self):
        super(ScanConnectedIntegTests, self).setUp()
        self.connection(_lv.PutItem(self.table, self.item))

    def tearDown(self):
        self.connection(_lv.DeleteItem(self.table, self.tab_key))
        super(ScanConnectedIntegTests, self).tearDown()

    def test_return_consumed_capacity_total(self):
        r = self.connection(_lv.Scan(self.table).return_consumed_capacity_total())

        self.assertEqual(r.consumed_capacity.capacity_units, 0.5)
        self.assertEqual(r.consumed_capacity.global_secondary_indexes, None)
        self.assertEqual(r.consumed_capacity.local_secondary_indexes, None)
        self.assertEqual(r.consumed_capacity.table, None)
        self.assertEqual(r.consumed_capacity.table_name, self.table)



# coding: utf8

# Copyright 2014-2015 Vincent Jacques <vincent@vincent-jacques.net>

"""
When given a :class:`Query`, the connection will return a :class:`QueryResponse`:

>>> connection(Query(table2).key_eq("h", 42))
<LowVoltage.actions.query.QueryResponse ...>

Items are accessed like this:

>>> connection(Query(table2).key_eq("h", 42)).items
[{u'h': 42, u'r1': 0, u'r2': 10}, {u'h': 42, u'r1': 1, u'r2': 9}, {u'h': 42, u'r1': 2, u'r2': 8}, {u'h': 42, u'r1': 3, u'r2': 7}, {u'h': 42, u'r1': 4, u'r2': 6}, {u'h': 42, u'r1': 5, u'r2': 5}, {u'h': 42, u'r1': 6}, {u'h': 42, u'r1': 7}, {u'h': 42, u'r1': 8}, {u'h': 42, u'r1': 9}]
"""

import LowVoltage as _lv
import LowVoltage.testing as _tst
from .action import Action
from .conversion import _convert_value_to_db, _convert_db_to_dict
from .next_gen_mixins import proxy
from .next_gen_mixins import (
    ConsistentRead,
    ExclusiveStartKey,
    ExpressionAttributeNames,
    ExpressionAttributeValues,
    FilterExpression,
    Limit,
    ProjectionExpression,
    ReturnConsumedCapacity,
    ScalarValue,
    Select,
)
from .return_types import ConsumedCapacity, _is_dict, _is_int, _is_list_of_dict


class QueryResponse(object):
    """
    The `Query response <http://docs.aws.amazon.com/amazondynamodb/latest/APIReference/API_Query.html#API_Query_ResponseElements>`__.
    """

    def __init__(
        self,
        ConsumedCapacity=None,
        Count=None,
        Items=None,
        LastEvaluatedKey=None,
        ScannedCount=None,
        **dummy
    ):
        self.__consumed_capacity = ConsumedCapacity
        self.__count = Count
        self.__items = Items
        self.__last_evaluated_key = LastEvaluatedKey
        self.__scanned_count = ScannedCount

    @property
    def consumed_capacity(self):
        """
        The capacity consumed by the request. If you used :meth:`~Query.return_consumed_capacity_total` or :meth:`~Query.return_consumed_capacity_indexes`.

        :type: ``None`` or :class:`.ConsumedCapacity`
        """
        if _is_dict(self.__consumed_capacity):  # pragma no branch (Defensive code)
            return ConsumedCapacity(**self.__consumed_capacity)

    @property
    def count(self):
        """
        The number of items matching the query.

        :type: ``None`` or long
        """
        if _is_int(self.__count):  # pragma no branch (Defensive code)
            return long(self.__count)

    @property
    def items(self):
        """
        The items matching the query. Unless you used :meth:`~Query.select_count`.

        :type: ``None`` or list of dict
        """
        if _is_list_of_dict(self.__items):  # pragma no branch (Defensive code)
            return [_convert_db_to_dict(i) for i in self.__items]

    @property
    def last_evaluated_key(self):
        """
        The key of the last item evaluated by the query. If not None, it should be given to :meth:`~Query.exclusive_start_key` is a subsequent :class:`Query`.

        :type: ``None`` or dict
        """
        if _is_dict(self.__last_evaluated_key):  # pragma no branch (Defensive code)
            return _convert_db_to_dict(self.__last_evaluated_key)

    @property
    def scanned_count(self):
        """
        The number of item scanned during the query. This can be different from :attr:`count` when using :meth:`~Query.filter_expression`.

        :type: ``None`` or long
        """
        if _is_int(self.__scanned_count):  # pragma no branch (Defensive code)
            return long(self.__scanned_count)


class Query(Action):
    """
    The `Query request <http://docs.aws.amazon.com/amazondynamodb/latest/APIReference/API_Query.html#API_Query_RequestParameters>`__.
    """

    def __init__(self, table_name):
        super(Query, self).__init__("Query", QueryResponse)
        self.__table_name = table_name
        self.__conditions = {}
        self.__consistent_read = ConsistentRead(self)
        self.__exclusive_start_key = ExclusiveStartKey(self)
        self.__expression_attribute_names = ExpressionAttributeNames(self)
        self.__expression_attribute_values = ExpressionAttributeValues(self)
        self.__filter_expression = FilterExpression(self)
        self.__index_name = ScalarValue("IndexName")(self)
        self.__limit = Limit(self)
        self.__projection_expression = ProjectionExpression(self)
        self.__return_consumed_capacity = ReturnConsumedCapacity(self)
        self.__scan_index_forward = ScalarValue("ScanIndexForward")(self)
        self.__select = Select(self)

    @property
    def payload(self):
        data = {"TableName": self.__table_name}
        if self.__conditions:
            data["KeyConditions"] = self.__conditions
        data.update(self.__consistent_read.payload)
        data.update(self.__exclusive_start_key.payload)
        data.update(self.__expression_attribute_names.payload)
        data.update(self.__expression_attribute_values.payload)
        data.update(self.__filter_expression.payload)
        data.update(self.__index_name.payload)
        data.update(self.__limit.payload)
        data.update(self.__projection_expression.payload)
        data.update(self.__return_consumed_capacity.payload)
        data.update(self.__scan_index_forward.payload)
        data.update(self.__select.payload)
        return data

    def key_eq(self, name, value):
        """
        Add a EQ condition to KeyConditions. Usable on both the hash key and the range key.
        The response will contain items whose key attribute ``name`` is equal to ``value``.

        >>> connection(Query(table2).key_eq("h", 42)).items
        [{u'h': 42, u'r1': 0, u'r2': 10}, {u'h': 42, u'r1': 1, u'r2': 9}, {u'h': 42, u'r1': 2, u'r2': 8}, ...]
        """
        self.__conditions[name] = {"ComparisonOperator": "EQ", "AttributeValueList": [_convert_value_to_db(value)]}
        return self

    def key_le(self, name, value):
        """
        Add a LE condition to KeyConditions. Usable only on the range key.
        The response will contain items whose key attribute ``name`` is less than or equal to ``value``.

        >>> connection(
        ...   Query(table2)
        ...     .key_eq("h", 42)
        ...     .key_le("r1", 1)
        ... ).items
        [{u'h': 42, u'r1': 0, u'r2': 10}, {u'h': 42, u'r1': 1, u'r2': 9}]
        """
        self.__conditions[name] = {"ComparisonOperator": "LE", "AttributeValueList": [_convert_value_to_db(value)]}
        return self

    def key_lt(self, name, value):
        """
        Add a LT condition to KeyConditions. Usable only on the range key.
        The response will contain items whose key attribute ``name`` is strictly less than ``value``.

        >>> connection(
        ...   Query(table2)
        ...     .key_eq("h", 42)
        ...     .key_lt("r1", 2)
        ... ).items
        [{u'h': 42, u'r1': 0, u'r2': 10}, {u'h': 42, u'r1': 1, u'r2': 9}]
        """
        self.__conditions[name] = {"ComparisonOperator": "LT", "AttributeValueList": [_convert_value_to_db(value)]}
        return self

    def key_ge(self, name, value):
        """
        Add a GE condition to KeyConditions. Usable only on the range key.
        The response will contain items whose key attribute ``name`` is greater than or equal to ``value``.

        >>> connection(
        ...   Query(table2)
        ...     .key_eq("h", 42)
        ...     .key_ge("r1", 7)
        ... ).items
        [{u'h': 42, u'r1': 7}, {u'h': 42, u'r1': 8}, {u'h': 42, u'r1': 9}]
        """
        self.__conditions[name] = {"ComparisonOperator": "GE", "AttributeValueList": [_convert_value_to_db(value)]}
        return self

    def key_gt(self, name, value):
        """
        Add a GT condition to KeyConditions. Usable only on the range key.
        The response will contain items whose key attribute ``name`` is strictly greater than ``value``.

        >>> connection(
        ...   Query(table2)
        ...     .key_eq("h", 42)
        ...     .key_gt("r1", 6)
        ... ).items
        [{u'h': 42, u'r1': 7}, {u'h': 42, u'r1': 8}, {u'h': 42, u'r1': 9}]
        """
        self.__conditions[name] = {"ComparisonOperator": "GT", "AttributeValueList": [_convert_value_to_db(value)]}
        return self

    def key_begins_with(self, name, value):
        """
        Add a BEGINS_WITH condition to KeyConditions. Usable only on the range key if it is a string.
        The response will contain items whose key attribute ``name`` begins with ``value``.
        """
        self.__conditions[name] = {"ComparisonOperator": "BEGINS_WITH", "AttributeValueList": [_convert_value_to_db(value)]}
        return self

    def key_between(self, name, lo, hi):
        """
        Add a BETWEEN condition to KeyConditions. Usable only on the range key.
        The response will contain items whose key attribute ``name`` is greater than or equal to ``lo`` and less than or equal to ``hi``.

        >>> connection(
        ...   Query(table2)
        ...     .key_eq("h", 42)
        ...     .key_between("r1", 4, 6)
        ... ).items
        [{u'h': 42, u'r1': 4, u'r2': 6}, {u'h': 42, u'r1': 5, u'r2': 5}, {u'h': 42, u'r1': 6}]
        """
        self.__conditions[name] = {"ComparisonOperator": "BETWEEN", "AttributeValueList": [_convert_value_to_db(lo), _convert_value_to_db(hi)]}
        return self

    @proxy("Query")
    def exclusive_start_key(self, key):
        """
        >>> r = connection(
        ...   Query(table2)
        ...     .key_eq("h", 42)
        ...     .limit(2)
        ... )
        >>> r.items
        [{u'h': 42, u'r1': 0, u'r2': 10}, {u'h': 42, u'r1': 1, u'r2': 9}]
        >>> r.last_evaluated_key
        {u'h': 42, u'r1': 1}
        >>> connection(
        ...   Query(table2)
        ...     .key_eq("h", 42)
        ...     .limit(2)
        ...     .exclusive_start_key({u'h': 42, u'r1': 1})
        ... ).items
        [{u'h': 42, u'r1': 2, u'r2': 8}, {u'h': 42, u'r1': 3, u'r2': 7}]
        """
        return self.__exclusive_start_key.set(key)

    @proxy
    def limit(self, limit):
        """
        See :meth:`exclusive_start_key` for an example.
        """
        return self.__limit.set(limit)

    @proxy
    def select_count(self):
        """
        >>> r = connection(
        ...   Query(table2)
        ...     .key_eq("h", 42)
        ...     .select_count()
        ... )
        >>> r.count
        10L
        >>> print r.items
        None
        """
        return self.__select.count()

    @proxy
    def select_all_attributes(self):
        """
        >>> connection(
        ...   Query(table2)
        ...     .key_eq("h", 42)
        ...     .index_name("lsi")
        ...     .select_all_attributes()
        ...     .limit(2)
        ... ).items
        [{u'h': 42, u'r1': 5, u'r2': 5}, {u'h': 42, u'r1': 4, u'r2': 6}]
        """
        return self.__select.all_attributes()

    @proxy
    def select_all_projected_attributes(self):
        """
        >>> connection(
        ...   Query(table2)
        ...     .key_eq("h", 42)
        ...     .index_name("lsi")
        ...     .select_all_projected_attributes()
        ...     .limit(2)
        ... ).items
        [{u'h': 42, u'r1': 5, u'r2': 5}, {u'h': 42, u'r1': 4, u'r2': 6}]
        """
        return self.__select.all_projected_attributes()

    def index_name(self, name):
        """
        Set Index. The request will use this index instead of the table key.

        >>> connection(
        ...   Query(table2)
        ...     .key_eq("h", 42)
        ...     .index_name("lsi")
        ... ).items
        [{u'h': 42, u'r1': 5, u'r2': 5}, {u'h': 42, u'r1': 4, u'r2': 6}, {u'h': 42, u'r1': 3, u'r2': 7}, {u'h': 42, u'r1': 2, u'r2': 8}, {u'h': 42, u'r1': 1, u'r2': 9}, {u'h': 42, u'r1': 0, u'r2': 10}]
        """
        return self.__index_name.set(name)

    def scan_index_forward_true(self):
        """
        Set ScanIndexForward to true. Items in the response will be sorted with ascending range keys.

        >>> connection(
        ...   Query(table2)
        ...     .key_eq("h", 42)
        ...     .project("r1")
        ...     .scan_index_forward_true()
        ... ).items
        [{u'r1': 0}, {u'r1': 1}, {u'r1': 2}, {u'r1': 3}, {u'r1': 4}, {u'r1': 5}, {u'r1': 6}, {u'r1': 7}, {u'r1': 8}, {u'r1': 9}]
        """
        return self.__scan_index_forward.set(True)

    def scan_index_forward_false(self):
        """
        Set ScanIndexForward to false. Items in the response will be sorted with descending range keys.

        >>> connection(
        ...   Query(table2)
        ...     .key_eq("h", 42)
        ...     .project("r1")
        ...     .scan_index_forward_false()
        ... ).items
        [{u'r1': 9}, {u'r1': 8}, {u'r1': 7}, {u'r1': 6}, {u'r1': 5}, {u'r1': 4}, {u'r1': 3}, {u'r1': 2}, {u'r1': 1}, {u'r1': 0}]
        """
        return self.__scan_index_forward.set(False)

    @proxy
    def project(self, *names):
        """
        >>> connection(
        ...   Query(table2)
        ...     .key_eq("h", 42)
        ...     .project("r2")
        ... ).items
        [{u'r2': 10}, {u'r2': 9}, {u'r2': 8}, {u'r2': 7}, {u'r2': 6}, {u'r2': 5}, {}, {}, {}, {}]
        """
        return self.__projection_expression.add(*names)

    @proxy
    def filter_expression(self, expression):
        """
        >>> connection(
        ...   Query(table2)
        ...     .key_eq("h", 42)
        ...     .key_ge("r1", 2)
        ...     .filter_expression("#syn IN (:val1, :val2)")
        ...     .expression_attribute_name("syn", "r2")
        ...     .expression_attribute_value("val1", 5)
        ...     .expression_attribute_value("val2", 7)
        ... ).items
        [{u'h': 42, u'r1': 3, u'r2': 7}, {u'h': 42, u'r1': 5, u'r2': 5}]
        """
        return self.__filter_expression.set(expression)

    @proxy
    def expression_attribute_name(self, synonym, name):
        """
        See :meth:`filter_expression` for an example.
        """
        return self.__expression_attribute_names.add(synonym, name)

    @proxy
    def expression_attribute_value(self, name, value):
        """
        See :meth:`filter_expression` for an example.
        """
        return self.__expression_attribute_values.add(name, value)

    @proxy
    def consistent_read_true(self):
        """
        >>> connection(
        ...   Query(table)
        ...     .key_eq("h", 0)
        ...     .consistent_read_true()
        ...     .return_consumed_capacity_total()
        ... ).consumed_capacity.capacity_units
        1.0
        """
        return self.__consistent_read.true()

    @proxy
    def consistent_read_false(self):
        """
        >>> connection(
        ...   Query(table)
        ...     .key_eq("h", 0)
        ...     .consistent_read_false()
        ...     .return_consumed_capacity_total()
        ... ).consumed_capacity.capacity_units
        0.5
        """
        return self.__consistent_read.false()

    @proxy
    def return_consumed_capacity_total(self):
        """
        >>> connection(
        ...   Query(table)
        ...     .key_eq("h", 0)
        ...     .return_consumed_capacity_total()
        ... ).consumed_capacity.capacity_units
        0.5
        """
        return self.__return_consumed_capacity.total()

    @proxy
    def return_consumed_capacity_indexes(self):
        """
        >>> c1 = connection(
        ...   Query(table)
        ...     .key_eq("h", 0)
        ...     .return_consumed_capacity_indexes()
        ... ).consumed_capacity
        >>> c1.capacity_units
        0.5
        >>> c1.table.capacity_units
        0.5
        >>> print c1.global_secondary_indexes
        None

        >>> c2 = connection(
        ...   Query(table)
        ...     .index_name("gsi")
        ...     .key_eq("gh", 0)
        ...     .return_consumed_capacity_indexes()
        ... ).consumed_capacity
        >>> c2.capacity_units
        0.5
        >>> c2.table.capacity_units
        0.0
        >>> c2.global_secondary_indexes["gsi"].capacity_units
        0.5
        """
        return self.__return_consumed_capacity.indexes()

    @proxy
    def return_consumed_capacity_none(self):
        """
        >>> print connection(
        ...   Query(table)
        ...     .key_eq("h", 0)
        ...     .return_consumed_capacity_none()
        ... ).consumed_capacity
        None
        """
        return self.__return_consumed_capacity.none()


class QueryUnitTests(_tst.UnitTests):
    def test_name(self):
        self.assertEqual(Query("Aaa").name, "Query")

    def test_table_name(self):
        self.assertEqual(Query("Aaa").payload, {"TableName": "Aaa"})

    def test_key_eq(self):
        self.assertEqual(
            Query("Aaa").key_eq("name", 42).payload,
            {
                "TableName": "Aaa",
                "KeyConditions": {"name": {"ComparisonOperator": "EQ", "AttributeValueList": [{"N": "42"}]}},
            }
        )

    def test_key_le(self):
        self.assertEqual(
            Query("Aaa").key_le("name", 42).payload,
            {
                "TableName": "Aaa",
                "KeyConditions": {"name": {"ComparisonOperator": "LE", "AttributeValueList": [{"N": "42"}]}},
            }
        )

    def test_key_lt(self):
        self.assertEqual(
            Query("Aaa").key_lt("name", 42).payload,
            {
                "TableName": "Aaa",
                "KeyConditions": {"name": {"ComparisonOperator": "LT", "AttributeValueList": [{"N": "42"}]}},
            }
        )

    def test_key_ge(self):
        self.assertEqual(
            Query("Aaa").key_ge("name", 42).payload,
            {
                "TableName": "Aaa",
                "KeyConditions": {"name": {"ComparisonOperator": "GE", "AttributeValueList": [{"N": "42"}]}},
            }
        )

    def test_key_gt(self):
        self.assertEqual(
            Query("Aaa").key_gt("name", 42).payload,
            {
                "TableName": "Aaa",
                "KeyConditions": {"name": {"ComparisonOperator": "GT", "AttributeValueList": [{"N": "42"}]}},
            }
        )

    def test_key_begins_with(self):
        self.assertEqual(
            Query("Aaa").key_begins_with("name", u"prefix").payload,
            {
                "TableName": "Aaa",
                "KeyConditions": {"name": {"ComparisonOperator": "BEGINS_WITH", "AttributeValueList": [{"S": "prefix"}]}},
            }
        )

    def test_key_between(self):
        self.assertEqual(
            Query("Aaa").key_between("name", 42, 44).payload,
            {
                "TableName": "Aaa",
                "KeyConditions": {"name": {"ComparisonOperator": "BETWEEN", "AttributeValueList": [{"N": "42"}, {"N": "44"}]}},
            }
        )

    def test_exclusive_start_key(self):
        self.assertEqual(Query("Aaa").exclusive_start_key({"h": u"v"}).payload, {"TableName": "Aaa", "ExclusiveStartKey": {"h": {"S": "v"}}})

    def test_limit(self):
        self.assertEqual(Query("Aaa").limit(4).payload, {"TableName": "Aaa", "Limit": 4})

    def test_select_all_attributes(self):
        self.assertEqual(Query("Aaa").select_all_attributes().payload, {"TableName": "Aaa", "Select": "ALL_ATTRIBUTES"})

    def test_select_all_projected_attributes(self):
        self.assertEqual(Query("Aaa").select_all_projected_attributes().payload, {"TableName": "Aaa", "Select": "ALL_PROJECTED_ATTRIBUTES"})

    def test_select_count(self):
        self.assertEqual(Query("Aaa").select_count().payload, {"TableName": "Aaa", "Select": "COUNT"})

    def test_expression_attribute_name(self):
        self.assertEqual(Query("Aaa").expression_attribute_name("n", "p").payload, {"TableName": "Aaa", "ExpressionAttributeNames": {"#n": "p"}})

    def test_expression_attribute_value(self):
        self.assertEqual(Query("Aaa").expression_attribute_value("n", u"p").payload, {"TableName": "Aaa", "ExpressionAttributeValues": {":n": {"S": "p"}}})

    def test_project(self):
        self.assertEqual(Query("Aaa").project("a").payload, {"TableName": "Aaa", "ProjectionExpression": "a"})

    def test_return_consumed_capacity_total(self):
        self.assertEqual(Query("Aaa").return_consumed_capacity_total().payload, {"TableName": "Aaa", "ReturnConsumedCapacity": "TOTAL"})

    def test_return_consumed_capacity_indexes(self):
        self.assertEqual(Query("Aaa").return_consumed_capacity_indexes().payload, {"TableName": "Aaa", "ReturnConsumedCapacity": "INDEXES"})

    def test_return_consumed_capacity_none(self):
        self.assertEqual(Query("Aaa").return_consumed_capacity_none().payload, {"TableName": "Aaa", "ReturnConsumedCapacity": "NONE"})

    def test_filter_expression(self):
        self.assertEqual(Query("Aaa").filter_expression("a=b").payload, {"TableName": "Aaa", "FilterExpression": "a=b"})

    def test_consistent_read_true(self):
        self.assertEqual(Query("Aaa").consistent_read_true().payload, {"TableName": "Aaa", "ConsistentRead": True})

    def test_consistent_read_false(self):
        self.assertEqual(Query("Aaa").consistent_read_false().payload, {"TableName": "Aaa", "ConsistentRead": False})

    def test_index_name(self):
        self.assertEqual(Query("Aaa").index_name("foo").payload, {"TableName": "Aaa", "IndexName": "foo"})

    def test_scan_index_forward_true(self):
        self.assertEqual(Query("Aaa").scan_index_forward_true().payload, {"TableName": "Aaa", "ScanIndexForward": True})

    def test_scan_index_forward_false(self):
        self.assertEqual(Query("Aaa").scan_index_forward_false().payload, {"TableName": "Aaa", "ScanIndexForward": False})


class QueryLocalIntegTests(_tst.LocalIntegTestsWithTableHR):
    def setUp(self):
        super(QueryLocalIntegTests, self).setUp()
        self.connection(_lv.BatchWriteItem().table("Aaa").put(
            {"h": u"0", "r": 41, "v": 0},
            {"h": u"0", "r": 42, "v": 1},
            {"h": u"0", "r": 43, "v": 2},
            {"h": u"0", "r": 44, "v": 3},
            {"h": u"0", "r": 45, "v": 4},
            {"h": u"1", "r": 42, "v": 2},
            {"h": u"2", "r": 42, "v": 3},
        ))

    def test_simple_query(self):
        r = self.connection(
            _lv.Query("Aaa").key_eq("h", u"1")
        )

        self.assertEqual(r.count, 1)
        self.assertEqual(r.items[0], {"h": "1", "r": 42, "v": 2})
        self.assertEqual(r.last_evaluated_key, None)
        self.assertEqual(r.scanned_count, 1)

    def test_complex_query(self):
        r = self.connection(
            _lv.Query("Aaa").key_eq("h", u"0").key_between("r", 42, 44)
                .scan_index_forward_false()
                .project("r", "v")
                .filter_expression("#p<>:v")
                .expression_attribute_name("p", "v")
                .expression_attribute_value("v", 2)
                .limit(2)
        )

        self.assertEqual(r.count, 1)
        self.assertEqual(r.items[0], {"r": 44, "v": 3})
        self.assertEqual(r.last_evaluated_key, {"h": u"0", "r": 43})
        self.assertEqual(r.scanned_count, 2)


class QueryConnectedIntegTests(_tst.ConnectedIntegTestsWithTable):
    def setUp(self):
        super(QueryConnectedIntegTests, self).setUp()
        self.connection(_lv.PutItem(self.table, self.item))

    def tearDown(self):
        self.connection(_lv.DeleteItem(self.table, self.tab_key))
        super(QueryConnectedIntegTests, self).tearDown()

    def test_return_consumed_capacity_total(self):
        r = self.connection(_lv.Query(self.table).key_eq("tab_h", u"0").return_consumed_capacity_total())

        self.assertEqual(r.consumed_capacity.capacity_units, 0.5)
        self.assertEqual(r.consumed_capacity.global_secondary_indexes, None)
        self.assertEqual(r.consumed_capacity.local_secondary_indexes, None)
        self.assertEqual(r.consumed_capacity.table, None)
        self.assertEqual(r.consumed_capacity.table_name, self.table)



# coding: utf8

# Copyright 2014-2015 Vincent Jacques <vincent@vincent-jacques.net>

import LowVoltage as _lv
import LowVoltage.testing as _tst
from .iterator import Iterator


class QueryIterator(Iterator):
    """Make as many "Query" actions as needed to iterate over all matching items."""

    def __init__(self, connection, query):
        Iterator.__init__(self, connection, query)

    def process(self, action, r):
        if r.last_evaluated_key is None:
            action = None
        else:
            action.exclusive_start_key(r.last_evaluated_key)
        return action, r.items


class QueryIteratorLocalIntegTests(_tst.LocalIntegTestsWithTableHR):
    keys = range(15)

    def setUp(self):
        super(QueryIteratorLocalIntegTests, self).setUp()
        self.connection(
            _lv.BatchWriteItem().table("Aaa").put(
                {"h": u"0", "r": r, "xs": "x" * 300000}  # 300kB items ensure a single Query will return at most 4 items
                for r in self.keys
            )
        )

    def test_simple_query(self):
        self.assertEqual(
            sorted(item["r"] for item in _lv.QueryIterator(self.connection, _lv.Query("Aaa").key_eq("h", u"0"))),
            self.keys
        )



# coding: utf8

# Copyright 2014-2015 Vincent Jacques <vincent@vincent-jacques.net>

import LowVoltage.testing as _tst
from LowVoltage.actions.tests.unit import *
from LowVoltage.compounds.tests.unit import *
from LowVoltage.connection.tests.unit import *


if __name__ == "__main__":  # pragma no branch (Test code)
    _tst.main()  # pragma no cover (Test code)



# coding: utf8

# Copyright 2014-2015 Vincent Jacques <vincent@vincent-jacques.net>

from ..connection import ConnectionUnitTests, SignerUnitTests, ResponderUnitTests
from ..credentials import Ec2RoleCredentialsUnitTests
from ..retry_policies import ExponentialBackoffRetryPolicyUnitTests



# coding: utf8

# Copyright 2014-2015 Vincent Jacques <vincent@vincent-jacques.net>

import LowVoltage as _lv


def global_setup():
    connection = _lv.Connection("us-west-2", _lv.EnvironmentCredentials())

    table1 = "LowVoltage.Tests.Doc.1"
    table2 = "LowVoltage.Tests.Doc.2"

    try:
        connection(_lv.DescribeTable(table1))
    except _lv.ResourceNotFoundException:
        connection(
            _lv.CreateTable(table)
                .hash_key("h", _lv.NUMBER).provisioned_throughput(1, 1)
                .global_secondary_index("gsi").hash_key("gh", _lv.NUMBER).range_key("gr", _lv.NUMBER).provisioned_throughput(1, 1).project_all()
        )

    try:
        connection(_lv.DescribeTable(table2))
    except _lv.ResourceNotFoundException:
        connection(
            _lv.CreateTable(table2)
                .hash_key("h", _lv.NUMBER).range_key("r1", _lv.NUMBER).provisioned_throughput(1, 1)
                .local_secondary_index("lsi").hash_key("h", _lv.NUMBER).range_key("r2", _lv.NUMBER).project_all()
        )

    _lv.WaitForTableActivation(connection, table1)
    _lv.BatchPutItem(
        connection,
        table1,
        [{"h": h, "gh": 0, "gr": 0} for h in range(10)],
    )

    _lv.WaitForTableActivation(connection, table2)
    _lv.BatchPutItem(
        connection,
        table2,
        [{"h": h, "r1": 0, "r2": 0} for h in range(10)],
    )

    _lv.BatchPutItem(
        connection,
        table2,
        [{"h": 42, "r1": r1, "r2": 10 - r1} for r1 in range(6)],
    )

    _lv.BatchPutItem(
        connection,
        table2,
        [{"h": 42, "r1": r1} for r1 in range(6, 10)],
    )

    return connection, table1, table2



#!/usr/bin/env python
# coding: utf8

# Copyright 2014-2015 Vincent Jacques <vincent@vincent-jacques.net>

import os
import setuptools

version = "0.5.0"


setuptools.setup(
    name="LowVoltage",
    version=version,
    description="Standalone DynamoDB client not hiding any feature",
    author="Vincent Jacques",
    author_email="vincent@vincent-jacques.net",
    url="http://pythonhosted.org/LowVoltage",
    packages=setuptools.find_packages(),
    license="MIT",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "License :: OSI Approved",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.4",
        "Environment :: Web Environment",
    ],
    install_requires=["requests>=2.2"],
    test_suite="LowVoltage.tests" if "AWS_ACCESS_KEY_ID" in os.environ else "LowVoltage.tests.local",
    test_loader="testresources:TestLoader",
    use_2to3=True,
    command_options={
        "build_sphinx": {
            "version": ("setup.py", version),
            "release": ("setup.py", version),
            "source_dir": ("setup.py", "doc"),
        },
    }
)



# coding: utf8

# Copyright 2014-2015 Vincent Jacques <vincent@vincent-jacques.net>

from ...wait_for_table_activation import WaitForTableActivationConnectedIntegTests
from ...wait_for_table_deletion import WaitForTableDeletionConnectedIntegTests
